"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


import numpy as np

from qrisp.misc import int_as_array, array_as_int
import sympy as sp
#from qiskit import *


#Class to describe truth tables
#Can be intialized with a list of bitstrings, or a numpy array with 1 and 0s or a sympy expression
class TruthTable:
    def __init__(self, init_object):
        
        #This handles the case that the initializing object is a sympy expression
        #the variables will be sorted alphabetically regarding their name
        if isinstance(init_object, sp.Expr):
            
            
            expr = init_object
            
            init_str = ""
            
            self.expr = expr
            from qrisp.arithmetic.poly_tools import get_ordered_symbol_list            
            #Retrieve a list of symbols in a given expression
            #Reverse to get the correct endian convention for qiskit
            symbol_list = get_ordered_symbol_list(expr)[::-1]
            
            #Replace the variables in the expression with 1s and 0s and add them to the intialization string
            from sympy import lambdify
            expr_func = lambdify([symbol_list], expr, "numpy")
            
            for i in range(2**len(symbol_list)):
                if len(symbol_list):
                    symbol_const = int_as_array(i, len(symbol_list))
                else:
                    symbol_const = []
                
                res = expr_func(symbol_const)%2
                if not set({int(res)}) <= set([0,1]):
                    raise Exception("Sympy expression returned non-boolean value")
                
                init_str += str(int(res))
            #this new object will now contain the string list which will be further prosecuted in this function
            init_object = [init_str]
        
        #This handles the case that the truth table is instantiated as a list of strings
        if isinstance(init_object, list):
            
            
            #Check if the list contains only bit strings of the same length
            for i in range(1, len(init_object)):
                if len(init_object[0]) != len(init_object[i]):
                    raise Exception("Tried to initialize truth table with varying lengths")
            
            #Save string list into object structure
            self.str_l = init_object
            
            #Generate numpy array representation
            self.n_rep = [[int(c) for c in singular_tt] for singular_tt in init_object]
            self.n_rep = np.array(self.n_rep).transpose()
        
        
        #This handles the case that the truth table is instantiated as a numpy array
        if isinstance(init_object, np.ndarray):
            
            #Save array into object structure
            self.n_rep = init_object
            
            #Generate list of bitstrings
            self.str_l = ["".join([str(int(init_object[i,j])) for i in range(init_object.shape[0])]) for j in range(init_object.shape[1])]
            
        
        #Check some conditions
        if not (set(self.n_rep.flatten()) <= set([0,1])):
            raise Exception("Tried to initialize truth table with a non boolean values")
            
        
        #Set shortcuts to some prevalent parameters of the truth table
        self.shape = self.n_rep.shape    
        
        
        if not (int(np.log2(self.shape[0])) == np.log2(self.shape[0])):
            raise Exception("Tried to initialize truth table with a boolean function with a length which is not an integer power of 2")
        
        
        
        self.bit_amount = int(np.log2(self.shape[0]))
        
    #Returns the "left side" of the truth table, ie. the variable values (starting with 00..00, 00..01, 00..10,...)
    def variable_array(self):
        variable_table_array = [int_as_array(i, int(np.log2(self.shape[0]))) for i in range(self.shape[0])]
        variable_table_array = np.array(variable_table_array)
            
        return TruthTable(variable_table_array)
    
    
    #Returns the cofactors of the truth table
    #If the truth table describes a boolean function f(x0,x1,x2...)
    #Then the cofactors for index 2 is f(x0,x1,0,...) and f(x0,x1,1,...)
    def cofactors(self, i):
        
        #Set up lists for the values of the cofactor tables
        cofactor_tables = [[],[]]
        
        #Iterate over all value constellations and put them either 
        #into the "first" or the "second" cofactor
        for k in range(self.shape[0]):
            if int_as_array(k, self.bit_amount)[i]:
                cofactor_tables[1].append(self.n_rep[k,:])
            else:
                cofactor_tables[0].append(self.n_rep[k,:])
        
        return [TruthTable(np.array(cofactor_tables[0])), TruthTable(np.array(cofactor_tables[1]))]
    
    #Returns a single column as a truth tables
    def sub_table(self, i):
        return TruthTable(np.array([self.n_rep[:,i]]).transpose())
    
    #Method to print truth tables
    def __str__(self):
        return str(np.array(self.n_rep))
    
    #Swaps to columns of the truth table
    def swap_col(self, i, j):
        temp_array = np.array(self.n_rep)
        temp_col = self.n_rep[:, i]
        temp_array[:, i] = temp_array[:, j]
        temp_array[:, j] = temp_col
        return TruthTable(temp_array)
    
    #Calculates the complexity of the truth table (more to that in the definition of "D")
    def calc_complexity(self, as_array = False):
        if as_array:
            return np.array([D(self.n_rep[:,i]) for i in range(len(self.str_l))])
        
        c = 0
        
        for i in range(len(self.str_l)):
            c += D(self.n_rep[:,i])
         
        return -c
    
            
    
    #Synthesizes a ciruict which represents the truth table
    def q_synth(self, input_var, output_var, method = "gray"):
        if output_var.size != self.shape[1]:
            raise Exception("Given output variable doesn't include the required amount of qubits")
        
        if len(input_var) != self.bit_amount:
            raise Exception("Given input variable doesn't include the required amount of qubits")
            
        #Use gray synthesis to synthesize truth table        
        if method == "gray":
            from qrisp.logic_synthesis.gray_synthesis import gray_logic_synth
                
            gray_logic_synth(input_var, output_var, self, phase_tolerant = False)
        
            
        #Use phase tolerant gray synthesis to synthesize truth table
        elif method == "gray_pt":
            from qrisp.logic_synthesis.gray_synthesis import gray_logic_synth
            
            gray_logic_synth(input_var, output_var, self, phase_tolerant = True)
        
        elif method == "gray_pt_inv":
            from qrisp.misc import quantum_invert
            from qrisp.logic_synthesis.gray_synthesis import gray_logic_synth
            
            quantum_invert(gray_logic_synth, [input_var, output_var, self, True], input_var.qs)
            
        elif method == "pprm_pt":
            from qrisp.logic_synthesis.pprm_synthesis import pprm
            
            pprm(input_var, output_var, self, phase_tolerant = True)
        
        elif method == "pprm":
            from qrisp.logic_synthesis.pprm_synthesis import pprm
            
            pprm(input_var, output_var, self, phase_tolerant = False)
            
        elif method == "best":
            input_var.qs.append(self.gate_synth(), list(input_var.reg) + list(output_var.reg))
        
        else:
            raise Exception("Given synthesis method unknown")
        
        
        #Append the truth table to the generated operation, so that it can be
        #resynthesized phase tolerantly during uncomputation
        # input_var.qs.data[-1].op.tt = self
        # input_var.qs.data[-1].op.logic_synth_method = method
        
        input_var.qs.data[-1].op = LogicSynthGate(input_var.qs.data[-1].op, self, phase_tolerant = method)
  
            
    def gate_synth(self, method = "best", inv = False):
        
        from qrisp.core import QuantumSession, QuantumVariable
        
            
        if method == "best":
            
            qs_list = []
            methods = ["gray_pt","td", "td_pk", "pprm_pt"]
            
            if self.bit_amount != 1:
                methods = ["gray_pt","pprm_pt"]
            else:
                methods = ["pprm_pt"]
            
            
            
            
            for m in methods:
                qs = QuantumSession()    
                input_var = QuantumVariable(self.bit_amount, qs)
                output_var = QuantumVariable(self.shape[1], qs)
                
                self.q_synth(input_var, output_var, method = m)
                
                qs_list.append(qs)
            
            cnot_list = [qs_list[i].cnot_count() for i in range(len(methods))]
            
            qs = qs_list[cnot_list.index(min(cnot_list))]
            
        else:         
            qs = QuantumSession()
            
            input_var = QuantumVariable(self.bit_amount, qs)
            output_var = QuantumVariable(self.shape[1], qs)
        
            self.q_synth(input_var, output_var, method = method)
            
        if not inv:
            return qs.to_gate()
        else:
            return qs.to_gate().inverse()
    
    def __or__(self, other):
        if self.shape != other.shape:
            raise Exception("Tried to or two truth tables of different shape")
            
        new_tt = np.zeros(self.shape)
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                new_tt[i,j] = (self.n_rep[i,j] or other.n_rep[i,j])
                
        return TruthTable(new_tt)
        
            
        
T_bib = {}
#Implementation of the truth table complexity measure described in Miller, D. M., “Spectral and Two-Place Decomposition Techniques in Reversible Logic,” Proc. Midwest Symposium on Circuits and Systems, on CD-ROM, August 2
def T(n):
    global T_bib
    
    if n in T_bib.keys():
        return T_bib[n]
    
    if n == 0:
        return np.array([[1]])
    
    result = np.zeros((2**n, 2**n))
    
    temp = T(n-1)
    
    result[:2**(n-1),:2**(n-1)] = temp
    result[2**(n-1):,:2**(n-1)] = temp
    result[:2**(n-1),2**(n-1):] = temp
    result[2**(n-1):,2**(n-1):] = -temp
    
    T_bib[n] = result
    return result


def fwht(a):
    """In-place Fast Walsh–Hadamard Transform of array a."""
    h = 1
    while h < len(a):
        for i in range(0, len(a), h * 2):
            for j in range(i, i + h):
                x = a[j]
                y = a[j + h]
                a[j] = x + y
                a[j + h] = x - y
        h *= 2

def rw_spectrum(f):
    
    if isinstance(f, str):
        f = [int(c) for c in f]
        f = np.array(f)
    
    a = f.copy()
    fwht(a)
    return a
    
    size = len(f)
    
    if np.log2(size) != int(np.log2(size)):
        raise Exception("The given function does not have the length to properly represent a truth table")
    
        
    n = int(np.log2(size))
    
    
    return np.dot(T(n), f)

def C(f):
    size = len(f)
    
    if np.log2(size) != int(np.log2(size)):
        raise Exception("The given function does not have the length to properly represent a truth table")
        
    n = int(np.log2(size))
    
    rw_spec = rw_spectrum(f)
    sum_ = 0
    for i in range(size):
        sum_ += sum(int_as_array(i, n))*rw_spec[i]**2
    
    sum_ = sum_/2**(n-2)
    
    return 1/2*(n*size - sum_)

def NZ(f):
    R = rw_spectrum(f)
    
    sum_ = 0
    for i in R:
        if i == 0:
            sum_ += 1
    return sum_


def D(f):
    size = len(f)
    if np.log2(size) != int(np.log2(size)):
        raise Exception("The given function does not have the length to properly represent a truth table")
        
    n = int(np.log2(size))
    
    
    return int(n*2**(n-3)*NZ(f) + C(f))


def synth_poly(truth_table, column = 0, coeff = None):
    
    if coeff == None:
        coeff = sp.symbols("".join([" x" + str(i) for i in range(truth_table.bit_amount)]))
        if truth_table.bit_amount == 1:
            coeff = [coeff]


    try:
        expr = truth_table.expr
        symbols = get_ordered_symbol_list(expr)
        coeff_temp = sp.symbols("".join([" abcdefghijkllmn" + str(i) for i in range(truth_table.bit_amount)]))
        subs_dic = {symbols[i]:coeff_temp[i] for i in range(len(coeff))}
        expr = expr.subs(subs_dic)
        subs_dic = {coeff_temp[i]:coeff[i] for i in range(len(coeff))}
        expr = expr.subs(subs_dic).expand()
        #print(subs_dic)
        #print(sp.Poly(expr.subs(subs_dic).expand(), domain = sp.GF(2)))
        return sp.Poly(expr.subs(subs_dic).expand(), domain = sp.GF(2)).expr
    except:
        pass

    
    poly = sp.sympify(0)
    for i in range(truth_table.shape[0]):
        if truth_table.n_rep[i,column]:
            temp = sp.sympify(1)
            array = int_as_array(i, truth_table.bit_amount)
            for j in range(len(array)):
                if array[j]:
                    temp *= coeff[-j-1]
                else:
                    temp *= coeff[-j-1]+1
            poly += temp
    
    if poly == 0:
        return poly
    return sp.Poly(poly.expand(), domain = sp.GF(2)).expr
    #return filter_pow(sp.Poly(poly.expand(), domain = sp.GF(2)).expr)
    
    
from qrisp.circuit import Operation
class LogicSynthGate(Operation):
    
    def __init__(self, init_op, tt, phase_tolerant = False):
        
        self.tt = tt
        
        self.logic_synth_method = phase_tolerant
        
        for var in init_op.__dict__.keys():
            if var != "inverse":
                self.__dict__[var] = init_op.__dict__[var]
                
    def inverse(self):
        return LogicSynthGate(Operation.inverse(self), self.tt, self.logic_synth_method)
    
    
def check_synthesis(tt, gate, log_output = False):
    synth_correct = True
    from qrisp.core import QuantumVariable, QuantumSession
    
    for i in range(tt.shape[0]):
        qs = QuantumSession()
        input_var = QuantumVariable(tt.bit_amount, qs)
        output_var = QuantumVariable(gate.num_qubits-tt.bit_amount, qs)
        int_encoder(input_var, i)
        qs.append(gate, range(gate.num_qubits))
        
        
        tt_value = "".join([str(x) for x in tt.n_rep[i]])
        synth_res = list(output_var.get_measurement().keys())[0]
        
        if log_output:
            print("TruthTable:", tt_value)
            print("Synthesis:", synth_res)
            print("---")
            
        synth_correct = synth_correct and tt_value == synth_res
    
    return synth_correct


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


import threading

from thrift.transport import TSocket
from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol
from qrisp.interface.thrift_interface.stoppable_thrift_server import StoppableThriftServer


# Returns the hosts ip
def get_ip():
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP


from qrisp.interface.thrift_interface.codegen import BackendService


# This class describes a Backend Server
class BackendServer:
    """
    This class allows convienent setup of a server respecting the Qrisp interface.
    
    Parameters
    ----------
    run_func : function
        A function that recieves a QuantumCircuit, an integer specifiying the amount of shots and a token in the form of a string. It returns the counts as a dictionary of bitstrings.
    socket_ip_address : str, optional
        The IP adress of where the listening socket should be opened. By default, the IP address of the hosting machine will be used.
    port : int, optional
        The port on which to listen for requests. By default 9090 will be used.
    name : str, optional
        A name for the server. The default is "generic_quantum_backend_server".
    ping_func : function, optional
        A function returning a BackendStatus object. The default is None.
    online_status : bool, optional
        A bool specifying wether the server should be able to handle requests directly after startup. The default is True.

    
    Examples
    --------
    
    We create a server listening on the localhost IP address using a run function which prints the token and queries the QASM-simulator. ::

        
        
        def run_func(qc, shots, token):
            
            #Convert to qiskit
            from qrisp.interface.circuit_converter import convert_circuit
            qiskit_qc = convert_circuit(qc, "qiskit")
            
            print(token)
            
            from qiskit import Aer
            qiskit_backend = Aer.get_backend('qasm_simulator')
            
            #Run Circuit on the Qiskit backend
            return qiskit_backend.run(qiskit_qc, shots = shots).result().get_counts()

        from qrisp.interface import BackendServer
        example_server = BackendServer(run_func, socket_ip_address = "127.0.0.1", port = 8080)
        example_server.start()

    
    """

    def __init__(self, run_func, socket_ip_address=None, port=None, transport=None,
                 name="generic_quantum_backend_server", ping_func=None, online_status=True):
    
        if socket_ip_address == None:
            socket_ip_address = get_ip()

        self.transport = transport

        self.socket_ip_address = socket_ip_address

        self.online_status = online_status

        self.run_func = run_func

        if ping_func == None:
            def ping_func():
                from qrisp.interface import BackendStatus
                return BackendStatus(name=name)

        self.ping_func = ping_func

        self.name = name

        self.port = port

    # Starts the server
    def start(self):
        """
        Starts the server.
        """

        if self.port == None:
            self.port = 9090

        self.thread_name = self.name + ":" + str(self.port)

        # Create the BackendServiceHandler class
        ping_func = self.ping_func
        run_func = self.run_func
        pass_online_status_bool_by_reference = lambda: self.online_status

        class BackendServiceHandler:

            def run(self, qc, shots, token):
                return run_func(qc, shots, token)

            def ping(self):
                status = ping_func()
                status.online = pass_online_status_bool_by_reference()
                return status

        # Create server
        handler = BackendServiceHandler()
        processor = BackendService.Processor(handler)
        tfactory = TTransport.TBufferedTransportFactory()

        if self.transport == None:
            self.transport = TSocket.TServerSocket(host=self.socket_ip_address, port=self.port)

        pfactory = TBinaryProtocol.TBinaryProtocolFactory()

        # server = TServer.TSimpleServer(processor, transport, tfactory, pfactory)
        # self.server = TServer.TThreadedServer(processor, self.transport, tfactory, pfactory, daemon = True)

        # self.server = TNonblockingServer.TNonblockingServer(processor, self.transport, tfactory, pfactory)
        self.server = StoppableThriftServer(processor, self.transport, tfactory, pfactory, daemon=True)

        # Create thread
        self.thr = threading.Thread(target=self.server.serve, name=self.thread_name)
        self.thr.setDaemon(True)

        # Start the thread
        self.thr.start()
        
        while True:
            if self.server.is_running:
                break

    def stop(self):
        """
        Stops the server.
        """
        if hasattr(self, "server"):
            self.server.stop()
            self.thr.join()

    def __del__(self):
        self.stop()


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


from qrisp.misc import bin_rep

def convert_circuit(qc, target_api = "thrift", transpile = True):

    
    if transpile or target_api == "open_api":
        from qrisp.circuit import transpile
        qc = transpile(qc)
    

    if target_api == "thrift":
        from qrisp.interface.thrift_interface.interface_types import PortableQuantumCircuit, PortableQubit, PortableClbit, PortableOperation, PortableInstruction
    elif target_api == "open_api":
        from qrisp.interface.openapi_interface.interface_types import PortableQuantumCircuit, PortableQubit, PortableClbit, PortableOperation, PortableInstruction
    elif target_api == "qrisp":
        return convert_to_qrisp(qc)
    elif target_api == "qiskit":
        return convert_to_qiskit(qc, transpile)
    

    
    
    circuit_qubits = [PortableQubit(identifier = qb.identifier) for qb in qc.qubits]    
    circuit_clbits = [PortableClbit(identifier = cb.identifier) for cb in qc.clbits]    
    
    
    
    res_qc = PortableQuantumCircuit(qubits = circuit_qubits, clbits = circuit_clbits, data = [])
    
    for i in range(len(qc.data)):
        instr = qc.data[i]
        if instr.op.name in ["qb_alloc", "qb_dealloc"]:
            # continue
            pass
        
        
        params = []
        for j in range(len(instr.op.params)):
            value = float(instr.op.params[j])
            params.append(value)
        
        
        if instr.op.definition != None:
            definition = convert_circuit(instr.op.definition, target_api, transpile = False)
            op = PortableOperation(name = instr.op.name, num_qubits = instr.op.num_qubits, num_clbits = instr.op.num_clbits, params = params, definition = definition)
        else:
            op = PortableOperation(name = instr.op.name, num_qubits = instr.op.num_qubits, num_clbits = instr.op.num_clbits, params = params)
            
        qubits = [PortableQubit(identifier = qb.identifier) for qb in instr.qubits]
        clbits = [PortableClbit(identifier = cb.identifier) for cb in instr.clbits]
        
        res_qc.data.append(PortableInstruction(op = op, qubits = qubits, clbits = clbits))
    
    return res_qc
    

def convert_to_qrisp(qc):
    from qrisp.circuit import QuantumCircuit, Qubit, Clbit, Operation, Instruction, op_list, fast_append
    
    op_dict = {op().name : op for op in op_list}
    
    res_qc = QuantumCircuit()
    
    
    #This dic gives the qiskit qubits/clbits when presented with their identifier
    bit_dic = {}
    
    #Add Qubits
    for i in range(len(qc.qubits)):
        res_qc.add_qubit(Qubit(qc.qubits[i].identifier))
        bit_dic[qc.qubits[i].identifier] = res_qc.qubits[-1]
        
    #Add Clbits
    for i in range(len(qc.clbits)):
        res_qc.add_clbit(Clbit(qc.clbits[i].identifier))
        bit_dic[qc.clbits[i].identifier] = res_qc.clbits[-1]
    
    
    with fast_append():
        for i in range(len(qc.data)):
            instr = qc.data[i]
            #Prepare qubits
            qubit_list = [bit_dic[qubit.identifier] for qubit in instr.qubits]
            clbit_list = [bit_dic[clbit.identifier] for clbit in instr.clbits]
    
            if not instr.op.definition is None:
                op = convert_to_qrisp(instr.op.definition).to_op()        
            elif instr.op.name in op_dict.keys():
                op = op_dict[instr.op.name](*instr.op.params)
            elif instr.op.name[5:] in op_dict.keys():
                if instr.op.name[:5] == "c_if_":
                    op = op_dict[instr.op.name[5:]](*instr.op.params).c_if()
            elif instr.op.name[-3:] == "_dg" and instr.op.name[:-3] in op_dict.keys():
                op = op_dict[instr.op.name[:-3]](*instr.op.params).inverse()
            else:
                raise Exception("Don't know how to convert operation " + str(instr.op.name) + " to qrisp")
            
            
            res_qc.data.append(Instruction(op, qubit_list, clbit_list))
        
    return res_qc
    
#Function to convert qrisp quantum circuits to Qiskit quantum circuits
def convert_to_qiskit(qc, transpile = False):
    
    from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister
        
    
    if transpile:
        from qrisp.circuit.transpiler import transpile
            
        qc = transpile(qc)
    
    #Create circuit

    qiskit_qc = QuantumCircuit()

    
    #This dic gives the qiskit qubits/clbits when presented with their identifier
    bit_dic = {}
    
    
    #Add Qubits
    for i in range(len(qc.qubits)):
        qiskit_qc.add_register(QuantumRegister(1, name = qc.qubits[i].identifier))
        
        bit_dic[qc.qubits[i].identifier] = qiskit_qc.qubits[-1]
        
    #Add Clbits
    for i in range(len(qc.clbits)):
        qiskit_qc.add_register(ClassicalRegister(1, name = qc.clbits[i].identifier))
        
        bit_dic[qc.clbits[i].identifier] = qiskit_qc.clbits[-1]
    

    #Add Instructions
    from qiskit.circuit import Instruction, ControlledGate, Barrier, Gate
    import qiskit.circuit.library.standard_gates as qsk_gates
    
    from qiskit.circuit.measure import Measure
    
    from qrisp.circuit import ControlledOperation
    for i in range(len(qc.data)):
        op = qc.data[i].op
        
        params = list(op.params)
        
        
        #Prepare qiskit qubits
        qubit_list = [bit_dic[qubit.identifier] for qubit in qc.data[i].qubits]
        clbit_list = [bit_dic[clbit.identifier] for clbit in qc.data[i].clbits]
        
        if hasattr(op, "global_phase"):
            qiskit_qc.global_phase += op.global_phase
        
        if op.name in ["qb_alloc", "qb_dealloc"]:
            continue
            temp = QuantumCircuit(1)
            qiskit_ins = temp.to_gate()
            qiskit_ins.name = op.name
            
        
        elif op.name == "gphase":
            temp_qc = QuantumCircuit(1)
            qiskit_ins = temp_qc.to_gate()
            qiskit_ins.name = "gphase"
        
        elif op.name == "cx":
            if hasattr(op, "ctrl_state"):
                qiskit_ins = qsk_gates.XGate().control(ctrl_state = op.ctrl_state)
            else:
                qiskit_ins = qsk_gates.XGate().control()
        elif op.name == "cz":
            if hasattr(op, "ctrl_state"):
                qiskit_ins = qsk_gates.ZGate().control(ctrl_state = op.ctrl_state)
            else:
                qiskit_ins = qsk_gates.ZGate().control()
        elif op.name == "cy":
            if hasattr(op, "ctrl_state"):
                qiskit_ins = qsk_gates.YGate().control(ctrl_state = op.ctrl_state)
            else:
                qiskit_ins = qsk_gates.YGate().control()
        
        elif issubclass(op.__class__, ControlledOperation):
            
            
            base_name = op.base_operation.name
            if len(base_name) == 1:
                base_name = base_name.upper()
            
            if op.base_operation.definition:
                
                qiskit_definition = convert_to_qiskit(op.base_operation.definition)
                base_gate = qiskit_definition.to_gate()
                base_gate.name = op.base_operation.name
                qiskit_ins = base_gate.control(len(op.controls), ctrl_state = op.ctrl_state[::-1])
                
            else:
                
                base_gate = create_qiskit_instruction(op.base_operation, params)
                qiskit_ins = base_gate.control(len(op.controls), ctrl_state = op.ctrl_state[::-1])
            
            
        else:
            qiskit_ins = create_qiskit_instruction(op, params)
            
        #Append to qiskit circuit
        qiskit_qc.append(qiskit_ins, qubit_list, clbit_list)
    
    
    #Return result
    return qiskit_qc

def create_qiskit_instruction(op, params = []):

    import qiskit.circuit.library.standard_gates as qsk_gates
    from qiskit.circuit.measure import Measure
    from qiskit.circuit import Barrier

    if op.name == "rxx":
        qiskit_ins = qsk_gates.RXXGate(*params)
    
    elif op.name == "measure":
        qiskit_ins = Measure()
    
    elif op.name == "swap":
        qiskit_ins = qsk_gates.SwapGate()
    
    elif op.definition:
        
        qiskit_definition = convert_to_qiskit(op.definition)
        if len(op.definition.clbits):
            qiskit_ins = qiskit_definition.to_instruction()
        else:
            qiskit_ins = qiskit_definition.to_gate()
        qiskit_ins.name = op.name
    
    elif op.name == "cp":
        qiskit_ins = qsk_gates.CPGate(params[0])
    elif op.name == "x":
        qiskit_ins = qsk_gates.XGate()
    elif op.name == "y":
        qiskit_ins = qsk_gates.YGate()
    elif op.name == "z":
        qiskit_ins = qsk_gates.ZGate()
    elif op.name == "rx":
        qiskit_ins = qsk_gates.RXGate(params[0])
    elif op.name == "ry":
        qiskit_ins = qsk_gates.RYGate(params[0])
    elif op.name == "rz":
        qiskit_ins = qsk_gates.RZGate(params[0])
    elif op.name == "sx":
        qiskit_ins = qsk_gates.SXGate()
    elif op.name == "h":
        qiskit_ins = qsk_gates.HGate()

    elif op.name == "p":
        qiskit_ins = qsk_gates.PhaseGate(params[0])
    elif op.name == "u1":
        qiskit_ins = qsk_gates.PhaseGate(params[0])
    elif op.name == "s":
        qiskit_ins = qsk_gates.SGate()
    elif op.name == "s_dg":
        qiskit_ins = qsk_gates.SGate().inverse()
    elif op.name == "t":
        qiskit_ins = qsk_gates.TGate()
    elif op.name == "t_dg":
        qiskit_ins = qsk_gates.TGate().inverse()
    elif op.name == "u3":
        qiskit_ins = qsk_gates.U3Gate(*params)
    elif op.name == "id":
        qiskit_ins = qsk_gates.IGate()
    elif op.name == "barrier":
        qiskit_ins = Barrier(op.num_qubits)
    else:
        raise Exception("Could not convert operation " + str(op.name) + " to Qiskit")
        
    return qiskit_ins


from qrisp.circuit.standard_operations import op_list
op_dic = {op().name : op for op in op_list}
op_dic["u"] = op_dic["u3"]

def convert_from_qiskit(qiskit_qc):
    
    from qiskit import transpile
    from qiskit.circuit import ControlledGate
    from qrisp import QuantumCircuit, Qubit, Clbit, op_list, ControlledOperation
    
    qc = QuantumCircuit()
    
    for i in range(qiskit_qc.num_qubits):
        q_reg = qiskit_qc.qubits[i]._register
        
        qubit_name = q_reg.name
        if q_reg.size > 1:
            qubit_name += "." + str(q_reg._bits.index(qiskit_qc.qubits[i]))
        
        qc.add_qubit(Qubit(qubit_name))
    
    for i in range(qiskit_qc.num_clbits):
        
        cl_reg = qiskit_qc.clbits[i]._register
        
        clbit_name = cl_reg.name
        if cl_reg.size > 1:
            clbit_name += "." + str(cl_reg._bits.index(qiskit_qc.clbits[i]))
        
        qc.add_clbit(Clbit(clbit_name))
    
    
    qb_dic = {qiskit_qc.qubits[i] : qc.qubits[i] for i in range(qiskit_qc.num_qubits)}
    cb_dic = {qiskit_qc.clbits[i] : qc.clbits[i] for i in range(qiskit_qc.num_clbits)}
    
    for i in range(len(qiskit_qc.data)):
        
        qiskit_op = qiskit_qc.data[i][0]
        
        if isinstance(qiskit_op, ControlledGate):
            controlled_gate = True
            qiskit_op = qiskit_op.base_gate
        else:
            controlled_gate = False
        
        
        params = list(qiskit_op.params)
        for j in range(len(qiskit_op.params)):
            if not isinstance(qiskit_op.params[j], float):                
                params[j] = float(qiskit_op.params[j])

    
        try:
            op = op_dic[qiskit_op.name](*params)
        except KeyError:
            try:    
                op = op_dic[qiskit_op.name.lower()](*params)
            except KeyError:
                if not qiskit_op.definition is None:
                    op = convert_from_qiskit(qiskit_op.definition).to_gate(name = qiskit_op.name)
                else:
                    raise Exception("Could not convert Qiskit operation " + str(qiskit_op.name) + " to Qrisp")
                

        if controlled_gate:
            qiskit_op = qiskit_qc.data[i][0]
            op = ControlledOperation(base_operation = op, num_ctrl_qubits = qiskit_op.num_ctrl_qubits, ctrl_state = bin_rep(qiskit_op.ctrl_state, qiskit_op.num_ctrl_qubits)[::-1])
        
        
        qubits = [qb_dic[qb] for qb in qiskit_qc.data[i][1]]
        clbits = [cb_dic[cb] for cb in qiskit_qc.data[i][2]]
        
        qc.append(op, qubits, clbits)
    
    return qc


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# -*- coding: utf-8 -*-

import numpy as np
from qrisp.simulator.bi_arrays import SparseBiArray, DenseBiArray, BiArray, tensordot
from numba import njit
import sympy as sp

id_matrix = np.eye(2, dtype = np.complex64)

pauli_x = np.asarray([[0,1],[1,0]], dtype = np.complex64)
pauli_y = np.asarray([[0,0],[0,0]]) +  1j*np.asarray([[0,-1],[1,0]])
pauli_z = np.asarray([[1,0],[0,-1]], dtype = np.complex64)


#Function which returns the unitary of a u3 gate
def u3matrix(theta, phi, lam, global_phase):
    
    import numpy as module
    res = module.zeros(shape = (2,2), dtype = module.complex64)
    
    from sympy.core.expr import Expr
    
    I = 1j
    for par in [theta, phi, lam, global_phase]:
        if isinstance(par, Expr):
            res = module.zeros(shape = (2,2), dtype = "object")
            import sympy as module
            I = sp.I
            break
    
    res[0,0] = module.cos(theta/2)
    res[0,1] = -module.exp(I*lam)*module.sin(theta/2)
    res[1,0] = module.exp(I*phi)*module.sin(theta/2)
    res[1,1] = module.exp(I*(phi+lam))*module.cos(theta/2)
    
    return res*module.exp(I*global_phase)
    
    
    
    

#Efficient function to generate the unitary of a controlled gate
def controlled_unitary(controlled_gate):
    
    
    m = controlled_gate.base_operation.num_qubits
    try:
        temp = controlled_gate.base_operation.unitary_array
    except AttributeError:
        temp = controlled_gate.base_operation.unitary_array = controlled_gate.base_operation.get_unitary()
    
    n = controlled_gate.num_qubits
    res = np.eye(2**n, dtype = temp.dtype)
    
    
    res[-2**m:,-2**m:] = temp
    
    controlled_gate.unitary = temp
    
    return res



    

#Calculates the unitary of a gate embeddded into a circuit of n qubits
#For instance consider this circuit
#Here we have gate = CXGate(), n = 4, destination_qubits = [2,3]
# q80_0: ─────
            
# q80_1: ─────
#        ┌───┐
# q80_2: ┤ X ├
#        └─┬─┘
# q80_3: ──■──
#This function then returns the overall unitary matrix in the form of either
#a BiArray or a numpy array depending what is more efficient
def calc_embedded_unitary(gate, n, destination_qubits):
    
    #First we generate the array corresponding to
    # 1 (x) 1 (x) ... gate
    #where 1 denotes the identity mapping on a qubit,
    # (x) the tensor product,
    # and gate the gate unitary

    #In principle we could use the numpy function np.kron    
    #tensor_array = np.kron(np.eye(2**(n-m)), gate.get_unitary())
    
    #However this performs alot of unnecessary multiplications
    #since the structure of this matrix is simply the unitary
    #of the gate on the diagonal 2**n/2**dim(gate) times repeated
    if not gate.definition is None:
        gate_array = __calc_circuit_unitary(gate.definition)
        gate.unitary = gate_array
    else:
        gate_array = gate.get_unitary()
    
    #Numpy arrays are only fast on a low scale than BiArrays. Therefore we generate
    #a Numpy Arrays
    if n < 10 and not gate_array.dtype == np.dtype("O"):
        tensor_array = generate_id_kron_jitted(gate_array, n)
    else:
        tensor_array = generate_id_kron(gate_array, n)
    
    #Reshape into an 2n-dimensional array where each axis represents an
    #index of the tensor factors
    tensor_array = tensor_array.reshape(n*[2,2])
    
    # #Generate the nec<essary swaps to put the destination qubits at the right place
    swaps = generate_swaps(destination_qubits, n)
    
    
    #Perform swaps
    for swap in swaps:
        tensor_array = swap_tensor_factors(tensor_array, *swap)
    

    #Shape back into the 2**n x 2**n matrix
    tensor_array = tensor_array.reshape([1<<n, 1<<n])
    
    
    if tensor_array.dtype == np.dtype("O") and n < 10:
        tensor_array = tensor_array.to_array()
    
    
    
    return tensor_array


#In order to generate the swaps we follow the strategy that 
#each destination qubit can be guaranteed to be puth in the correct place
#with a single swap. Therefore we iterate through the destination qubits and
#successively put them in the right place
def generate_swaps(destination_qubits, n):
    
    #This list represents the current state of the tensor
    #The list element represents an identity tensor
    tensor_state = (n - len(destination_qubits))*["id"] + list(destination_qubits)
    swaps = []
    
    #Function to check if the tensor state reached its desired state
    def check_if_reached(tensor_state):
        for i in range(len(tensor_state)):
            if tensor_state[i] != "id":
                if tensor_state[i] != i:
                    return False
        return True

    
    #Repat until the tensor reached its final state
    while not check_if_reached(tensor_state):
        
        #Iterate through the tensor state and find a qubit which is not
        #in the correct position.
        for i in range(len(tensor_state)):
            if tensor_state[i] != "id":
                
                #If tensor_state[i] == i the qubit in question already has reached the final state
                if tensor_state[i] != i:
                    
                    #Swap the list elements
                    index = int(i)
                    value = int(tensor_state[i])
                    tensor_state[index], tensor_state[value] = tensor_state[value], tensor_state[index]
                    
                    #Log the swaps
                    swaps.append((index, value))
                    
                    #Restart search for qubit in an incorrect position
                    break

    return list(swaps)




#Swaps the axis of a kroneker product (tensor product of linear mappings)
#ie. if we insert tensor = 1 (x) A (x) 1 (x) 1
#and apply this function with i = 0, j = 1
#we get A (x) 1 (x) 1 (x) 1
def swap_tensor_factors(tensor, i, j):
    n = len(tensor.shape)//2
    
    #Note that we have to swap two axes because we are 
    #dealing with the kroneker product and not the tensor product
    
    tensor = tensor.swapaxes(i, j)
    tensor = tensor.swapaxes(n+i,n + j)
    
    return tensor


#In principle we could use the numpy function np.kron    
#tensor_array = np.kron(np.eye(2**(n-m)), gate.get_unitary())

#However this performs alot of unnecessary multiplications
#since the structure of this matrix is simply the unitary
#of the gate on the diagonal 2**n/2**dim(gate) times repeated
def generate_id_kron(input_tensor, n):
    
    if isinstance(input_tensor, np.ndarray):
        if input_tensor.dtype == np.dtype("O"):
            input_tensor = DenseBiArray(input_tensor)
        else:
            input_tensor = SparseBiArray(input_tensor)
            
    k = int(np.log2(input_tensor.shape[0]))
    
    data = np.ones(2**(n-k))
    
    nz_indices = np.arange(2**(n-k)).astype(np.int64)
    nz_indices = (nz_indices*2**(n-k) + nz_indices)
    
    id_tensor = SparseBiArray((nz_indices, data), shape = (2**(n-k), 2**(n-k)))  
    
    id_tensor.reshape((n-k)*[2,2])
    input_tensor.reshape(k*[2,2])
    
    res = tensordot(id_tensor, input_tensor, ((),()))
    
    
    for i in range(n-k):
        for j in range(k):
            res.swapaxes(2*n - 2*k -1 + j - i, 2*n - 2*k + j -i)
    
    
    
    
    res.reshape([2**n, 2**n])
    
    return res
    
@njit(cache = True)
def generate_id_kron_jitted(input_tensor, n):
    
    res = np.zeros((2**n, 2**n), dtype = input_tensor.dtype)
    
    m = int(input_tensor.shape[0])
    for i in range(2**n//m):
        res[i*m:(i+1)*m,i*m:(i+1)*m] = input_tensor
        
    return res


#This functions calculates the unitary of a given circuit

#The idea to make this more efficient is that not all matrix multiplication have to be perforned
#in the 2**n dimensional basis

#For instance in this circuit:

#      ┌───┐     
# q_0: ┤ X ├──■──
#      └───┘┌─┴─┐
# q_1: ─────┤ X ├
#           └───┘
# q_2: ──────────
               
# q_3: ──────────

#We can calculate the unitary on the first two qubits on the 4 d basis
#and then embedd this unitary into the 2**4 = 16d basis


#In order to harness the effciency gain we use a divide and conquer strategy:
#We merge the instruction of the circuit into pairs of elementary gates, calculate
#this unitary and the again merge this unitary with the neighbouring pair of elementary gate
#Applying this recursively the requires only a single 2**n d matrix multiplication

def __calc_circuit_unitary(qc):

    n = len(qc.qubits)

    #If the circuit is empty, return the identity matrix    
    if len(qc.data) == 0:
        return np.eye(2**n)
    
    #If the circuit contains only a single insturction,
    #calculate this instructions unitary and embedd it
    if len(qc.data) == 1:
        
        instr_0 = qc.data[0]
        unitary_0 = calc_embedded_unitary(qc.data[0].op, n, [qc.qubits.index(qb) for qb in instr_0.qubits])
        
        return unitary_0
    
    
    #If the circuit is a pair of instructions and uses all qubits,
    #we calculate their product by matrix multiplication
    
    #If the circuit contained unused qubits, the next "else" statement will turn
    #this circuit into it's version without any idle qubits
    #and this case is then called at a later recursion
    elif len(qc.data) == 2 and len(qc.qubits) == len(set(qc.data[0].qubits).union(qc.data[1].qubits)):
            
        instr_0 = qc.data[0]    
        instr_1 = qc.data[1]
        
        unitary_0 = calc_embedded_unitary(instr_0.op, n, [qc.qubits.index(qb) for qb in instr_0.qubits])
        unitary_1 = calc_embedded_unitary(instr_1.op, n, [qc.qubits.index(qb) for qb in instr_1.qubits])
        
        res = tensordot(unitary_1, unitary_0, (1,0), contract_sparsity_threshold = 0.01)
        
    #This is the case where we generate the pairs of instructions
    else:
        
        #Set up a new circuit where the merged instructions are appended
        recursion_qc = qc.clearcopy()
        
        #Merge every pair of neighbouring instructions into one instruction
        for i in range(0, len(qc.data), 2):
            if i+1 < len(qc.data):
                recursion_qc.append(qc.data[i].merge(qc.data[i+1]))
            else:
                recursion_qc.append(qc.data[i])
        #The recursion circuit now has about half of the instructions of the original quantum circuit
        #Therefore the recursion is finite
        res = __calc_circuit_unitary(recursion_qc)
    
            
    return res    

from qrisp import fast_append
#The result of __calc_circuit_unitary can either be a numpy array or a BiArray.
#This is important in order for the recursion structure to work out
#This function is a wrapper that converts to the desired output
def calc_circuit_unitary(qc, res_type = "numpy"):
    with fast_append():
        res = __calc_circuit_unitary(qc)
    
    if res_type == "numpy":
        
        if isinstance(res, BiArray):
            res = res.to_array()
        
        return res
    
    elif res_type == "bi_array":
        
        if isinstance(res, BiArray):
            return res
        
        return DenseBiArray(res)
        
    
    else:
        raise Exception("Don't know result type " + res_type)


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


def outome_array_hashing_test():
    from qrisp import OutcomeArray
    import numpy as np
    
    test = OutcomeArray(np.eye(3))
    
    test_dic = {test : 1}
    
    test_dic[str(np.eye(3))]
    
    assert str(np.eye(3)) == str(test)



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 27.04.2022
import pytest

from qrisp.logic_synthesis import TruthTable, gray_logic_synth
from qrisp.misc import int_encoder
from qrisp.core import QuantumVariable, QuantumSession

@pytest.mark.parametrize("num", ["001", "011", "000", "100"])
def test_gray_synthesis_example(num):
    # split the input string into a list of characters
    tmp = list(num)

    # Create a testing truth table
    tt = TruthTable(tmp)

    # Create new quantum session
    qs = QuantumSession()

    # Create input variable
    input_var = QuantumVariable(tt.bit_amount, qs)

    # Encode index

    int_encoder(input_var, 0)

    # Create output variable
    output_var = QuantumVariable(tt.shape[1], qs)

    # Perform logic synthesis
    gray_logic_synth(input_var, output_var, tt, phase_tolerant=False)

    # Expected bitstring comes from tt in format array [0 0 1]
    # create string for assert comparison
    ttstr = ''.join([str(item) for item in tt.n_rep[0, :]])
    mes_res = output_var.get_measurement()
    
    assert ttstr == str(list(mes_res.keys())[0])
    assert ttstr in ['001', '011', '000', '100'] and str(list(mes_res.keys())[0]) in ['001', '011', '000', '100']






"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


path = str(__file__)[:-11]
client_path = path + "/codegen/client"
server_path = path + "/codegen/server"

import sys
sys.path.insert(0, client_path)
sys.path.insert(0, server_path)

from qrisp.interface.openapi_interface.backend_client import BackendClient
from qrisp.interface.openapi_interface.backend_server import BackendServer
from qrisp.interface.openapi_interface.interface_types import *


sys.path.pop(0)
sys.path.pop(0)



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


import numpy as np
from qrisp.circuit import Operation


class GXX_wrapper(Operation):
    
    def __init__(self, n, chi):
        
        if n == 1:
            pass
            #raise Exception("Tried to create 1D GXX-gate")
        flattened_chi = sum(chi, [])
        
        super().__init__(name = "GXX", num_qubits = n, params = flattened_chi)
        
        self.chi = chi
        self.n = n
        from qrisp.circuit import QuantumCircuit
        qc = QuantumCircuit(n)
        
        for i in range(len(chi)):
            for j in range(len(chi[i])):
                if chi[i][j] == 0 or i == j:
                    continue
                qc.rxx(chi[i][j]/2, i, j)
        
        self.definition = qc
        
    def inverse(self):
        chi_inv = [[-x for x in y] for y in self.chi]
        return GXX_wrapper(self.n, chi_inv)
    

# Mimics a circuit which performs n cphase gates, where each cp interacts with the last qubit
# example
# q2874_0: ─■─────────────────────────────────────────────────────────────
#           │                                                             
# q2874_1: ─┼────────■────────────────────────────────────────────────────
#           │        │                                                    
# q2874_2: ─┼────────┼────────■───────────────────────────────────────────
#           │        │        │                                           
# q2874_3: ─┼────────┼────────┼────────■──────────────────────────────────
#           │        │        │        │                                  
# q2874_4: ─┼────────┼────────┼────────┼────────■─────────────────────────
#           │        │        │        │        │                         
# q2874_5: ─┼────────┼────────┼────────┼────────┼────────■────────────────
#           │        │        │        │        │        │                
# q2874_6: ─┼────────┼────────┼────────┼────────┼────────┼────────■───────
#           │P(π/2)  │P(π/2)  │P(π/2)  │P(π/2)  │P(π/2)  │P(π/2)  │P(π/2) 
# q2874_7: ─■────────■────────■────────■────────■────────■────────■───────
# Using only one GMS gate or two uniform GMS gates
def gms_multi_cp_gate_mono_phase(n, theta, use_uniform = True, phase_tolerant = False, basis = 'GXX'):
    
    from qrisp import QuantumSession, QuantumVariable, h, cp, p
    qs = QuantumSession()
    qv = QuantumVariable(n+1, qs)
    if basis == 'GXX':
        if use_uniform:
            gms_1 = GXX_wrapper((n+1), (n+1)*[(n+1)*[-theta/2]])
            gms_2 = GXX_wrapper(n, n*[n*[theta/2]])
        else:
            #gms = GXX_wrapper(n, (n+1)*[n*[0] + [-theta/2]])
            for i in range(n):
                cp(theta, qv[i], qv[n])
                # qs.cp(theta, i, n)
            return GXX_converter(qs).to_gate()

        if not phase_tolerant:
            for i in range(n):
                p(theta/2, qv[i])
                # qs.p(theta/2, qv.reg[i])

        p(theta/2*n, qv[-1])
        # qs.p(theta/2*n, qv.reg[-1])

        h(qv)

        qs.append(gms_1, qv.reg)
        if not phase_tolerant:
            qs.append(gms_2, qv.reg[:-1])

        h(qv)


    elif basis == 'GZZ':
        if use_uniform:
            gms_1 = GZZ_wrapper((n+1), (n+1)*[(n+1)*[-theta/2]])
            gms_2 = GZZ_wrapper(n, n*[n*[theta/2]])
        else:
            #gms = GXX_wrapper(n, (n+1)*[n*[0] + [-theta/2]])
            for i in range(n):
                cp(theta, qv[i], qv[n])
                # qs.cp(theta, i, n)
            return GZZ_converter(qs).to_gate()

        if not phase_tolerant:
            for i in range(n):
                p(theta/2, qv[i])
                # qs.p(theta/2, qv.reg[i])

        p(theta/2*n, qv[-1])
        # qs.p(theta/2*n, qv.reg[-1])

        qs.append(gms_1, qv.reg)
        if not phase_tolerant:
            qs.append(gms_2, qv.reg[:-1])

    else:
        raise 'Basis choice must be GXX or GZZ'

    return qs.to_gate(name = "gms_mono_phase_cp")


# qb_0: ──■─────────────────
#         │                 
# qb_1: ──┼────■────────────
#         │    │            
# qb_2: ──┼────┼────■───────
#         │    │    │       
# qb_3: ──┼────┼────┼────■──
#       ┌─┴─┐┌─┴─┐┌─┴─┐┌─┴─┐
# qb_4: ┤ X ├┤ X ├┤ X ├┤ X ├
#       └───┘└───┘└───┘└───┘
def gms_multi_cx_fan_in(n, use_uniform = True, phase_tolerant = False, basis = 'GXX'):
    
    from qrisp import QuantumSession, QuantumVariable, h
    qs = QuantumSession()
    qv = QuantumVariable(n+1, qs)
    
    
    h(qv[-1])
    
    qs.append(gms_multi_cp_gate_mono_phase(n, np.pi, use_uniform = use_uniform, phase_tolerant = phase_tolerant, basis = basis), qv.reg)
    
    h(qv[-1])
    
    #result = qs.to_gate(name = "gms_multi_cx_fan_in")
    #result.name = "gms_multi_cx_fan_in"
    
    return qs.to_gate(name = "gms_multi_cx_fan_in")


# qubit_0: ──■────■────■────■──
#          ┌─┴─┐  │    │    │  
# qubit_1: ┤ x ├──┼────┼────┼──
#          └───┘┌─┴─┐  │    │  
# qubit_2: ─────┤ x ├──┼────┼──
#               └───┘┌─┴─┐  │  
# qubit_3: ──────────┤ x ├──┼──
#                    └───┘┌─┴─┐
# qubit_4: ───────────────┤ x ├
#                         └───┘
def gms_multi_cx_fan_out(n, use_uniform = True, phase_tolerant = False, basis = 'GXX'):
    
    from qrisp import QuantumSession, QuantumVariable, h
    qs = QuantumSession()
    qv = QuantumVariable(n+1, qs)
    
    for i in range(n):
        h(qv[i])
    
    qs.append(gms_multi_cp_gate_mono_phase(n, np.pi, use_uniform = use_uniform, phase_tolerant = phase_tolerant, basis = basis), qv.reg)
    
    for i in range(n):
        h(qv[i])
    
    
    result = qs.to_gate()
    result.name = "gms_multi_cx_fan_out"
    
    return result


#This functions takes a circuit which only consists of phase and cphase gates
#and turns it into a GXX gate + some one-qubit gates
#The basic idea is that a cphase gate acts as
#CP |ab> = exp(i theta/2 (1 - (-1)^(ab))) |ab>
#We rewrite (-1)^(ab) = -1/2*((-1)^(a+b) - ((-1)^a + (-1)^b) - 1)
#The term (-1)^(a+b) can now be executed by a Moelmer-Soerensen:
#exp(i theta/2 Z_a (x) Z_b)
#The terms (-1)^a and (-1)^b are executed by single qubit phase gates
#The remaining -1 is an irrelevant global phase

def GXX_converter(qs):
    
    
    from qrisp import QuantumSession, QuantumVariable
    
    
    # Check if the given Circuit is valid
    global_phase = np.array([0.])
    for i in range(len(qs.data)):
        if not qs.data[i].op.name in ["p", "cp", "id", "cz", "rz", "qb_alloc", "qb_dealloc"]:
            raise Exception(qs.data[i].op.name + " is neither Phase nor CPhase gate")
        
        if not qs.data[i].op.name in ["cp", "cz", "qb_alloc", "qb_dealloc"]:
            global_phase += qs.data[i].op.global_phase
        
        
        # Subtract a 1/4 of the phase in case we are dealing with a cp gate (for more elaboration, check eq. 99 of https://arxiv.org/abs/2112.10537)
        if qs.data[i].op.name == "cp":
            global_phase -= qs.data[i].op.params[0]/4
            
        elif qs.data[i].op.name == "cz":
            global_phase -= np.pi/4

    # Create qubit index dictionary
    qubit_index_dic = {qs.qubits[i] : i for i in range(len(qs.qubits))}

    # Create phase matrix
    # The (i,j) entry of this matrix contains the phase that is applied onto the qubit pair (i,j)
    # the indices with i == j represent single qubit rz gates
    n = len(qs.qubits)
    phase_matrix = np.zeros((n,n))
    
    # Go through all data entries of the circuit
    for i in range(len(qs.data)):

        # Set instruction alias
        ins = qs.data[i]
        
        if ins.op.name in ["id", "qb_alloc", "qb_dealloc"]:
            continue
        
        # If sinlge qubit gate => Collect phase
        if len(ins.qubits) == 1:
            index = qubit_index_dic[ins.qubits[0]]

            phase_matrix[index, index] += ins.op.params[0]
            continue
        
        
        #Get indices
        index_0 = qubit_index_dic[ins.qubits[0]]
        index_1 = qubit_index_dic[ins.qubits[1]]
        
        #Swap indices if neccessary such that the phase matrix is an upper triangle matrix
        if index_0 >= index_1:
            temp = int(index_0)
            index_0 = index_1
            index_1 = temp
        
        
        #Collect phase
        if ins.op.name == "cz":
            phase_matrix[index_0, index_1] += np.pi
        else:
            phase_matrix[index_0, index_1] += ins.op.params[0]
    
    
    #Create Quantum Session
    qc_res = qs.clearcopy()
    
    
    if global_phase[0] != 0:
        qc_res.gphase(global_phase[0], 0)
        
    
    #Calculate the row-sum and the column sum of the phase matrix
    #to determine which single qubit phase has to be applied
    #This is because every entry with the same row / column represents a phase gate
    #where the qubit in question participated
    for i in range(n):
        qc_res.p((sum(phase_matrix[i,:]) + sum(phase_matrix[:,i]))/2, i)
        phase_matrix[i,i] = 0
    
    
    #Prepary Chi list for GXX gate
    chi_list = [[-phase_matrix[j,i] for i in range(n)] for j in range(n)]
    
    #Apply GXX Gate
    qc_res.h(qc_res.qubits)
    qc_res.append(GXX_wrapper(n, chi_list), range(n))
    qc_res.h(qc_res.qubits)
    
    return qc_res




#Similar to gms_multi_cp_gate_mono_phase but also allows more than one phase
#ie. converts circuits of the type 
# q54264_0: ─■─────────────────────────────────────────
#            │                                         
# q54264_1: ─┼─────────■───────────────────────────────
#            │         │                               
# q54264_2: ─┼─────────┼──────────■────────────────────
#            │         │          │                    
# q54264_3: ─┼─────────┼──────────┼──────────■─────────
#            │P(-π/5)  │P(-2π/5)  │P(-3π/5)  │P(-4π/5) 
# q54264_4: ─■─────────■──────────■──────────■─────────
#The advantage over the GXX_converter is the use_uniform mode, which makes
#sure only uniform GMS gates are used
def gms_multi_cp_gate(n, phases, use_uniform = True, basis = 'GXX'):
    
    from qrisp import QuantumSession, QuantumVariable, p
    
    qs = QuantumSession()
    
    qv = QuantumVariable(n+1, qs)
    
    
    for i in range(n):
        p(phases[i]/2, qv[i])
        # qs.p(phases[i]/2, i)
    
    p(sum(phases)/2, qv[n])
    # qs.p(sum(phases)/2, n)
    
    qs.append(gms_multi_cx_fan_out(n, use_uniform = use_uniform, phase_tolerant = False, basis = basis), qv.reg)
    
    for i in range(n):
        qs.p(-phases[i]/2, i)
    
    qs.append(gms_multi_cx_fan_out(n, use_uniform = use_uniform, phase_tolerant = False, basis = basis).inverse(), qv.reg)
    
    result = qs.to_gate()
    
    result.name = "gms_multi_cp_gate"
    
    return result


class GZZ_wrapper(Operation):

    def __init__(self, n, chi):

        if n == 1:
            pass
            # raise Exception("Tried to create 1D GZZ-gate")
        flattened_chi = sum(chi, [])

        super().__init__(name="GZZ", num_qubits=n, params=flattened_chi)

        self.chi = chi
        self.n = n
        from qrisp.circuit import QuantumCircuit
        qc = QuantumCircuit(n)

        for i in range(len(chi)):
            for j in range(len(chi[i])):
                if chi[i][j] == 0 or i == j:
                    continue
                qc.rzz(chi[i][j] / 2, i, j)
        
        self.definition = qc


#This functions takes a circuit which only consists of phase and cphase gates
#and turns it into a GZZ gate + some one-qubit gates
#The basic idea is that a cphase gate acts as
#CP |ab> = exp(i theta/2 (1 - (-1)^(ab))) |ab>
#We rewrite (-1)^(ab) = -1/2*((-1)^(a+b) - ((-1)^a + (-1)^b) - 1)
#The term (-1)^(a+b) can now be executed by a Moelmer-Soerensen:
#exp(i theta/2 Z_a (x) Z_b)
#The terms (-1)^a and (-1)^b are executed by single qubit phase gates
#The remaining -1 is an irrelevant global phase
def GZZ_converter(qs):
    from qrisp import QuantumSession, QuantumVariable

    # Check if the given Circuit is valid
    global_phase = np.array([0.])
    for i in range(len(qs.data)):
        if not qs.data[i].op.name in ["p", "cp", "id", "cz", "rz", "qb_alloc", "qb_dealloc"]:
            raise Exception(qs.data[i].op.name + " is neither Phase nor CPhase gate")
        
        if not qs.data[i].op.name in ["cp", "cz", "qb_alloc", "qb_dealloc"]:
            global_phase += qs.data[i].op.global_phase

        # Subtract a 1/4 of the phase in case we are dealing with a cp gate (for more elaboration, check eq. 99 of https://arxiv.org/abs/2112.10537)
        if qs.data[i].op.name == "cp":
            global_phase -= qs.data[i].op.params[0]/4
            
        elif qs.data[i].op.name == "cz":
            global_phase -= np.pi/4

    # Create qubit index dictionary
    qubit_index_dic = {qs.qubits[i]: i for i in range(len(qs.qubits))}

    # Create phase matrix
    # The (i,j) entry of this matrix contains the phase that is applied onto the qubit pair (i,j)
    # the indices with i == j represent single qubit rz gates
    n = len(qs.qubits)
    phase_matrix = np.zeros((n, n))

    # Go through all data entries of the circuit
    for i in range(len(qs.data)):

        # Set instruction alias
        ins = qs.data[i]

        if ins.op.name in ["id", "qb_alloc", "qb_dealloc"]:
            continue

        # If sinlge qubit gate => Collect phase
        if len(ins.qubits) == 1:
            index = qubit_index_dic[ins.qubits[0]]

            phase_matrix[index, index] += ins.op.params[0]
            continue

        # Get indices
        index_0 = qubit_index_dic[ins.qubits[0]]
        index_1 = qubit_index_dic[ins.qubits[1]]

        # Swap indices if neccessary such that the phase matrix is an upper triangle matrix
        if index_0 >= index_1:
            temp = int(index_0)
            index_0 = index_1
            index_1 = temp

        # Collect phase
        if ins.op.name == "cz":
            phase_matrix[index_0, index_1] += np.pi
        else:
            phase_matrix[index_0, index_1] += ins.op.params[0]

    # Create Quantum Session
    qc_res = qs.clearcopy()
    
    if global_phase[0] != 0:
        qc_res.gphase(global_phase[0], 0)
        
    # Calculate the row-sum and the column sum of the phase matrix
    # to determine which single qubit phase has to be applied
    # This is because every entry with the same row / column represents a phase gate
    # where the qubit in question participated
    for i in range(n):
        qc_res.p((sum(phase_matrix[i, :]) + sum(phase_matrix[:, i])) / 2, i)
        phase_matrix[i, i] = 0

    # Prepary Chi list for GZZ gate
    chi_list = [[-phase_matrix[j, i] for i in range(n)] for j in range(n)]

    # Apply GZZ Gate
    qc_res.append(GZZ_wrapper(n, chi_list), range(n))

    return qc_res


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

#
# Autogenerated by Thrift Compiler (0.15.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
all_structs = []


class Qubit(object):
    """
    Attributes:
     - identifier

    """


    def __init__(self, identifier=None,):
        self.identifier = identifier

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.identifier = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Qubit')
        if self.identifier is not None:
            oprot.writeFieldBegin('identifier', TType.STRING, 1)
            oprot.writeString(self.identifier.encode('utf-8') if sys.version_info[0] == 2 else self.identifier)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Clbit(object):
    """
    Attributes:
     - identifier

    """


    def __init__(self, identifier=None,):
        self.identifier = identifier

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.identifier = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Clbit')
        if self.identifier is not None:
            oprot.writeFieldBegin('identifier', TType.STRING, 1)
            oprot.writeString(self.identifier.encode('utf-8') if sys.version_info[0] == 2 else self.identifier)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Operation(object):
    """
    Attributes:
     - name
     - num_qubits
     - num_clbits
     - params
     - definition

    """


    def __init__(self, name=None, num_qubits=None, num_clbits=None, params=None, definition=None,):
        self.name = name
        self.num_qubits = num_qubits
        self.num_clbits = num_clbits
        self.params = params
        self.definition = definition

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRING:
                    self.name = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.I32:
                    self.num_qubits = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.I32:
                    self.num_clbits = iprot.readI32()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.LIST:
                    self.params = []
                    (_etype3, _size0) = iprot.readListBegin()
                    for _i4 in range(_size0):
                        _elem5 = iprot.readDouble()
                        self.params.append(_elem5)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 5:
                if ftype == TType.STRUCT:
                    self.definition = QuantumCircuit()
                    self.definition.read(iprot)
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Operation')
        if self.name is not None:
            oprot.writeFieldBegin('name', TType.STRING, 1)
            oprot.writeString(self.name.encode('utf-8') if sys.version_info[0] == 2 else self.name)
            oprot.writeFieldEnd()
        if self.num_qubits is not None:
            oprot.writeFieldBegin('num_qubits', TType.I32, 2)
            oprot.writeI32(self.num_qubits)
            oprot.writeFieldEnd()
        if self.num_clbits is not None:
            oprot.writeFieldBegin('num_clbits', TType.I32, 3)
            oprot.writeI32(self.num_clbits)
            oprot.writeFieldEnd()
        if self.params is not None:
            oprot.writeFieldBegin('params', TType.LIST, 4)
            oprot.writeListBegin(TType.DOUBLE, len(self.params))
            for iter6 in self.params:
                oprot.writeDouble(iter6)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.definition is not None:
            oprot.writeFieldBegin('definition', TType.STRUCT, 5)
            self.definition.write(oprot)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Instruction(object):
    """
    Attributes:
     - op
     - qubits
     - clbits

    """


    def __init__(self, op=None, qubits=None, clbits=None,):
        self.op = op
        self.qubits = qubits
        self.clbits = clbits

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.STRUCT:
                    self.op = Operation()
                    self.op.read(iprot)
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.qubits = []
                    (_etype10, _size7) = iprot.readListBegin()
                    for _i11 in range(_size7):
                        _elem12 = Qubit()
                        _elem12.read(iprot)
                        self.qubits.append(_elem12)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.LIST:
                    self.clbits = []
                    (_etype16, _size13) = iprot.readListBegin()
                    for _i17 in range(_size13):
                        _elem18 = Clbit()
                        _elem18.read(iprot)
                        self.clbits.append(_elem18)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Instruction')
        if self.op is not None:
            oprot.writeFieldBegin('op', TType.STRUCT, 1)
            self.op.write(oprot)
            oprot.writeFieldEnd()
        if self.qubits is not None:
            oprot.writeFieldBegin('qubits', TType.LIST, 2)
            oprot.writeListBegin(TType.STRUCT, len(self.qubits))
            for iter19 in self.qubits:
                iter19.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.clbits is not None:
            oprot.writeFieldBegin('clbits', TType.LIST, 3)
            oprot.writeListBegin(TType.STRUCT, len(self.clbits))
            for iter20 in self.clbits:
                iter20.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class QuantumCircuit(object):
    """
    Attributes:
     - qubits
     - clbits
     - data
     - init

    """


    def __init__(self, qubits=None, clbits=None, data=None, init=True,):
        self.qubits = qubits
        self.clbits = clbits
        self.data = data
        self.init = init

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.qubits = []
                    (_etype24, _size21) = iprot.readListBegin()
                    for _i25 in range(_size21):
                        _elem26 = Qubit()
                        _elem26.read(iprot)
                        self.qubits.append(_elem26)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.LIST:
                    self.clbits = []
                    (_etype30, _size27) = iprot.readListBegin()
                    for _i31 in range(_size27):
                        _elem32 = Clbit()
                        _elem32.read(iprot)
                        self.clbits.append(_elem32)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 3:
                if ftype == TType.LIST:
                    self.data = []
                    (_etype36, _size33) = iprot.readListBegin()
                    for _i37 in range(_size33):
                        _elem38 = Instruction()
                        _elem38.read(iprot)
                        self.data.append(_elem38)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 4:
                if ftype == TType.BOOL:
                    self.init = iprot.readBool()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('QuantumCircuit')
        if self.qubits is not None:
            oprot.writeFieldBegin('qubits', TType.LIST, 1)
            oprot.writeListBegin(TType.STRUCT, len(self.qubits))
            for iter39 in self.qubits:
                iter39.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.clbits is not None:
            oprot.writeFieldBegin('clbits', TType.LIST, 2)
            oprot.writeListBegin(TType.STRUCT, len(self.clbits))
            for iter40 in self.clbits:
                iter40.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.data is not None:
            oprot.writeFieldBegin('data', TType.LIST, 3)
            oprot.writeListBegin(TType.STRUCT, len(self.data))
            for iter41 in self.data:
                iter41.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.init is not None:
            oprot.writeFieldBegin('init', TType.BOOL, 4)
            oprot.writeBool(self.init)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Message(object):
    """
    Attributes:
     - qubits
     - annotation

    """


    def __init__(self, qubits=None, annotation=None,):
        self.qubits = qubits
        self.annotation = annotation

    def read(self, iprot):
        if iprot._fast_decode is not None and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None:
            iprot._fast_decode(self, iprot, [self.__class__, self.thrift_spec])
            return
        iprot.readStructBegin()
        while True:
            (fname, ftype, fid) = iprot.readFieldBegin()
            if ftype == TType.STOP:
                break
            if fid == 1:
                if ftype == TType.LIST:
                    self.qubits = []
                    (_etype45, _size42) = iprot.readListBegin()
                    for _i46 in range(_size42):
                        _elem47 = Qubit()
                        _elem47.read(iprot)
                        self.qubits.append(_elem47)
                    iprot.readListEnd()
                else:
                    iprot.skip(ftype)
            elif fid == 2:
                if ftype == TType.STRING:
                    self.annotation = iprot.readString().decode('utf-8', errors='replace') if sys.version_info[0] == 2 else iprot.readString()
                else:
                    iprot.skip(ftype)
            else:
                iprot.skip(ftype)
            iprot.readFieldEnd()
        iprot.readStructEnd()

    def write(self, oprot):
        if oprot._fast_encode is not None and self.thrift_spec is not None:
            oprot.trans.write(oprot._fast_encode(self, [self.__class__, self.thrift_spec]))
            return
        oprot.writeStructBegin('Message')
        if self.qubits is not None:
            oprot.writeFieldBegin('qubits', TType.LIST, 1)
            oprot.writeListBegin(TType.STRUCT, len(self.qubits))
            for iter48 in self.qubits:
                iter48.write(oprot)
            oprot.writeListEnd()
            oprot.writeFieldEnd()
        if self.annotation is not None:
            oprot.writeFieldBegin('annotation', TType.STRING, 2)
            oprot.writeString(self.annotation.encode('utf-8') if sys.version_info[0] == 2 else self.annotation)
            oprot.writeFieldEnd()
        oprot.writeFieldStop()
        oprot.writeStructEnd()

    def validate(self):
        return

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(Qubit)
Qubit.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'identifier', 'UTF8', None, ),  # 1
)
all_structs.append(Clbit)
Clbit.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'identifier', 'UTF8', None, ),  # 1
)
all_structs.append(Operation)
Operation.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'name', 'UTF8', None, ),  # 1
    (2, TType.I32, 'num_qubits', None, None, ),  # 2
    (3, TType.I32, 'num_clbits', None, None, ),  # 3
    (4, TType.LIST, 'params', (TType.DOUBLE, None, False), None, ),  # 4
    (5, TType.STRUCT, 'definition', [QuantumCircuit, None], None, ),  # 5
)
all_structs.append(Instruction)
Instruction.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'op', [Operation, None], None, ),  # 1
    (2, TType.LIST, 'qubits', (TType.STRUCT, [Qubit, None], False), None, ),  # 2
    (3, TType.LIST, 'clbits', (TType.STRUCT, [Clbit, None], False), None, ),  # 3
)
all_structs.append(QuantumCircuit)
QuantumCircuit.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'qubits', (TType.STRUCT, [Qubit, None], False), None, ),  # 1
    (2, TType.LIST, 'clbits', (TType.STRUCT, [Clbit, None], False), None, ),  # 2
    (3, TType.LIST, 'data', (TType.STRUCT, [Instruction, None], False), None, ),  # 3
    (4, TType.BOOL, 'init', None, True, ),  # 4
)
all_structs.append(Message)
Message.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'qubits', (TType.STRUCT, [Qubit, None], False), None, ),  # 1
    (2, TType.STRING, 'annotation', 'UTF8', None, ),  # 2
)
fix_spec(all_structs)
del all_structs



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# -*- coding: utf-8 -*-

#The functions in this file are supposed to reorder circuits in such a way,
#that measurements, resets and disentanglers are performed as early as possible.
#This is beneficial because the simulation of two decoherent states can be parallelized better
#furthermore might the measurement of these imply that one of the states has vanishing probability
#which means that this state does not require further simulation.

#The idea to achieve the ordering is to convert the circuit into a directed acyclic graph (now called causal graph),
#where two successive operations with overlapping qubits are represented with two nodes,
#connected by a directed edge. The directed edge, however points into the opposite direction
#of the sequential order of the operations. This way, if we want to evaluate
#what gates are necessary in order to perform a measurement, disentangling or reset,
#we simply have to look at the set of gates that is reachable from that specific node in the causal graph

#Consider the following circuit
#           ┌───┐        
#  qubit_9: ┤ X ├──■─────
#           ├───┤┌─┴─┐┌─┐
# qubit_10: ┤ Y ├┤ x ├┤M├
#           ├───┤└───┘└╥┘
# qubit_11: ┤ H ├──────╫─
#           └───┘      ║ 
#  clbit_0: ═══════════╩═
                      
#In order to perform the measurement, we have to execute the x and the cx gate
#The corresponding causal graph
#(measure)-> (CX)->(X)
#               L->(Y)
#(H)
#We see that CX, X and Y are reachable from the measurement
#We now perform the perform a depth first traversal on the
#subgraph of the nodes which are reachable from the measurement.

#It is important to node that for a regular depth first traversal,
#we are free to choose which child of a node to traverse next.
#This is not the casehere, because this might disturb the topological
#order, which would imply that we performed an illegitimat reordering
#Therefore, in this depth first traversal, we always traverse the child
#with the highest index in a topological order first.

import networkx as nx
import numpy as np

def nx_get_causal_graph(qc, inverted = False, get_non_unitary_nodes = False, preferential_gates = []):
    from qrisp.circuit import Operation, Instruction
    
    #Create graph object
    G = nx.DiGraph()
    
    #This distionary contains the information, which node of the graph
    #if the most up to date noce for a qubit
    current_node_qubits = {}
    
    #List to collect the non-unitary nodes
    non_unitary_nodes = []
    
    #Now we traverse the circuit instructions
    for i in range(len(qc.data)):
        
        #Add new node to the graph
        new_node = i
        G.add_node(new_node)
        
        #Determine the set of nodes, which the new node will connect to
        #ie. the nodes which are most up to date on the particular qubits
        node_set = []
        for qb in qc.data[i].qubits:
            
            #If there is a node which has been connected to this qubit before,
            #append the node to the node set, that the new node will be connected to
            try:
                node_set.append(current_node_qubits[qb])
            except KeyError:
                pass
            
            #Update the dictionary
            current_node_qubits[qb] = new_node
            
        #Make sure every node is listed only once
        node_set = list(set(node_set))
        
        #Add the edges
        if inverted:
            for node in node_set:
                G.add_edge(new_node, node)
        else:
            for node in node_set:
                G.add_edge(node, new_node)
        
        #Log if the new node is non unitary
        if qc.data[i].op.name in preferential_gates + ["final_op"]:
            non_unitary_nodes.append(new_node)
        
        
    #Return result
    if get_non_unitary_nodes:
        return G, non_unitary_nodes
    
    return G


#Function to reorder a circuit as described above
def nx_reorder_circuit(qc, preferential_gates = []):
    
    from qrisp.circuit import Operation, Instruction
    
    for i in range(len(qc.qubits)):
        qc.append(Operation("final_op", num_qubits = 1), [qc.qubits[i]])
    
    #Acquire causal graph
    G, non_unitary_nodes = nx_get_causal_graph(qc, inverted = True, get_non_unitary_nodes = True, preferential_gates = preferential_gates)
    
    from networkx import descendants, topological_sort, transitive_reduction
    
    #We now order the non-unitary nodes according to how many decendents they all have
    #Measurements/Resets/Disentanglings with only a few descendants need only a few
    #gates to be simulated until the measurement can be executed
    node_costs = []
    for i in range(len(non_unitary_nodes)):
        if qc.data[hash(non_unitary_nodes[i])].op.name == "final_op":
            node_costs.append((non_unitary_nodes[i], np.inf))
        else:
            node_costs.append((non_unitary_nodes[i], len(descendants(G, non_unitary_nodes[i]))))
    
    node_costs.sort(key = lambda x : x[1])
    
    #We now determine the topological dictionary of the nodes
    #This dictionary assigns each node an integer, which indicates it's
    #position in a topological ordering. The topological ordering of a graph
    #has the feature that two nodes N1, N2 which stand in a causal relationship to each other
    #(ie. N1 has to be executed after N2) also have increasing index in the topological
    #ordering ie. 
    #tp_dic[N1] > tp_dic[N2]
    sorted_nodes = list(topological_sort(G))
    tp_dic = {sorted_nodes[i] : i for i in range(len(sorted_nodes))}

    #This function now performs depth first traversal of the given causal graph,
    #starting at node. Each time a node is visited the callback function is called
    
    #The callback function that this will mainly be used with, is a logger,
    #which notes in which order the nodes have been visited
    
    #As mentioned in the comment at the beginning of this file,
    #a regular depth first traversal basically allows picking any child
    #to traverse next. In our case, this can mess with the topological ordering
    #Therefore we allways traverse the child with the highest topological index
    
    def topological_desc_traversal(G, node, tp_dic, callback):
        
        node_list = [x for x in nx.descendants(G, node)]
        
        node_list.sort(key = lambda x : -tp_dic[x])
        
        for n in node_list:
            callback(n)
            G.remove_node(n)
            
        callback(node)
        G.remove_node(node)
        return
        
    def topological_df_traversal(G, node, tp_dic, callback):
        #Acquire list of children
        node_list = [x for x in G.neighbors(node)]
        
        #Sort according to their topological index
        node_list.sort(key = lambda x : -tp_dic[x])
        
        #Recursively traverse the children
        while node_list:
            topological_df_traversal(G, node_list[0], tp_dic, callback)
            
            #Since the traversed nodes are removed after each traversal,
            #we need to to update the list of children, in case any of
            #the nodes in node_list have been removed
            node_list = [x for x in G.neighbors(node)]
            node_list.sort(key = lambda x : -tp_dic[x])
            
        #Call the callback function
        callback(node)
        #Remove the node in order to prevent traversing it again
        #this can happen because the causal graph is not necessarily a tree graph
        G.remove_node(node)
    
    
    
    
    #The circuits in this list will be the circuits whose
    #execution is the absolute minimum in order to evaluate a certain non-unitary operation
    new_qc_list = []
    
    #Now we succesively determine the minimal circuit required to execute
    #for each non-unitary operation
    while node_costs:
        
        #This node contains the non-unitary operation
        evaluation_node = node_costs.pop(0)[0]
        
        #This list will contain the unitary operations that are necessary
        #in order to perform the operation described by evaluation_node
        evaluation_list = []
        
        #Create callback function
        def callback(x):
            evaluation_list.append(x)
        
        #Traverse causal graph
        topological_desc_traversal(G, evaluation_node, tp_dic, callback)
        # topological_df_traversal(G, evaluation_node, tp_dic, callback)
        
        
        #Create circuit
        new_qc = qc.clearcopy()    
        #Append the corresponding instruction to the circuit
        for node in evaluation_list:
            instr = qc.data[hash(node)]
            if not instr.op.name == "final_op":
                new_qc.data.append(instr)
                
        new_qc_list.append(new_qc)
        
    
    #Create result quantum circuit
    new_qc = qc.clearcopy()
    
    #Concatenate the data of the newly created circuits
    new_qc.data = sum([qc.data for qc in new_qc_list], [])
    
    
    #Remove final_op operations
    for i in range(len(qc.qubits)):
        qc.data.pop(-1)
    
    #Return result    
    return new_qc


#Similar function as above but implemented for the C++ based
#graph theory package networkit
def nk_reorder_circuit(qc, preferential_gates = []):
    
    from qrisp.circuit import Operation, Instruction
    
    for i in range(len(qc.qubits)):
        qc.append(Operation("final_op", num_qubits = 1), [qc.qubits[i]])
        
    import networkit as nk
    G = nk.Graph(directed = True)
    current_node_qubits = {qubit : "-" for qubit in qc.qubits}
    non_unitary_nodes = []
    measurement_counter = 0
    for i in range(len(qc.data)):
        new_node = G.addNode()
        node_set = []
        for qb in qc.data[i].qubits:
            if current_node_qubits[qb] != "-":
                node_set.append(current_node_qubits[qb])
                
            current_node_qubits[qb] = new_node
        node_set = list(set(node_set))
        
        for node in node_set:
            G.addEdge(i, node)
            
        if qc.data[i].op.name in preferential_gates + ["final_op"]:
            non_unitary_nodes.append(new_node)
            if qc.data[i].op.name == "measure":
                measurement_counter += 1

    
    
    from networkit.traversal import Traversal
    
    new_qc_list = []
    
    from networkit.graphtools import GraphTools
    sorted_nodes = GraphTools.topologicalSort(G)

    tp_dic = {sorted_nodes[i] : i for i in range(len(sorted_nodes))}


    reach_alg = nk.reachability.ReachableNodes(G, exact = True)
    
    reach_alg.run()
    node_costs = []
    


    
    for i in range(len(non_unitary_nodes)):
        if qc.data[hash(non_unitary_nodes[i])].op.name == "final_op":
            node_costs.append((non_unitary_nodes[i], np.inf))
        else:
            node_costs.append((non_unitary_nodes[i], reach_alg.numberOfReachableNodes(non_unitary_nodes[i])))
    
    node_costs.sort(key = lambda x : x[1])
    
    def topological_desc_traversal(G, node, tp_dic, callback):
        
        node_list = []
        
        nk.traversal.Traversal.DFSfrom(G, node, node_list.append)
        
        node_list.sort(key = lambda x : -tp_dic[x])
        
        for n in node_list:
            callback(n)
            G.removeNode(node)
        
        # callback(node)
        # G.removeNode(node)    
        
        return
        
        
    
    def topological_df_traversal(G, node, tp_dic, callback):
        
        node_list = [x for x in G.iterNeighbors(node)]
        node_list.sort(key = lambda x : -tp_dic[x])
        
        while node_list:
            topological_df_traversal(G, node_list[0], tp_dic, callback)
            node_list = [x for x in G.iterNeighbors(node)]
            node_list.sort(key = lambda x : -tp_dic[x])
            
        
        callback(node)
        G.removeNode(node)
        
    
    while node_costs:
            
            
        evaluation_node = node_costs.pop(0)[0]
        
        evaluation_list = []
        
        def callback(x, y = 0):
            
            if len(evaluation_list):
                G.removeNode(evaluation_list[-1])
            evaluation_list.append(x)
            
        
        new_qc = qc.clearcopy()    
        
        
        # topological_df_traversal(G, evaluation_node, tp_dic, callback)
        topological_desc_traversal(G, evaluation_node, tp_dic, callback)
            
        for node in evaluation_list:
            instr = qc.data[hash(node)]
            if instr.op.name != "final_op":
                new_qc.data.append(instr)
    
        new_qc_list.append(new_qc)
    
    for i in range(len(qc.qubits)):
        qc.data.pop(-1)    
    
    new_qc = qc.clearcopy()
    new_qc.data = sum([qc.data for qc in new_qc_list], [])
    
    return new_qc

try:
    import networkit
    nk_available = True
    
except:
    nk_available = False
    # print("Install networkit for additional simulator performance. Resorting to networkx.")
    


def reorder_circuit(qc, preferential_gates = []):
    if nk_available:
        # return nk_reorder_circuit(qc, preferential_gates)
    
        return nx_reorder_circuit(qc, preferential_gates)
    else:
        return nx_reorder_circuit(qc, preferential_gates)
    


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 29.04.2022
import pytest
import numpy

from qrisp.interface import VirtualQiskitBackend
from qrisp import QuantumFloat, QuantumChar, QuantumArray, merge, x, h, OutcomeArray, multi_measurement


# check if the respone from QuantumArray.get_measurement() has the expected formet.
# Should return a list of tuples of the type (numpy.ndarray, float)
# ie. [(array([1,1,0]), 232), (array([1,1,3]), 115), ...]
def test_conditional_environments_example():
    
    qc = QuantumChar(name = "q ch")

    qc[:] = "f"

    qf = QuantumFloat(4,-1, signed = True)
    q_array = QuantumArray(qf, 4)


    h(qc[0])

    q_array[0] += -0.5


    with "f" == qc:
        h(q_array[0][0])
        
        
        with q_array[0] == -0.5:
            q_array[2].encode(-0.5)
            q_array[1] -= -1
            with q_array[1] == 1:
                q_array[3].init_from(q_array[2])
        
        #Test in-environment printing
        print(qc.qs)


    print(multi_measurement([qc, q_array]))
    assert multi_measurement([qc, q_array]) == {('e', OutcomeArray([-0.5,  0. ,  0. ,  0. ])): 0.5, ('f', OutcomeArray([-1.,  0.,  0.,  0.])): 0.25, ('f', OutcomeArray([-0.5,  1. , -0.5, -0.5])): 0.25}
    
    assert len(qc.qs.qv_list) == 5

    from qrisp import as_hamiltonian, QuantumBool, control
    @as_hamiltonian
    def dummy_d_function(i, j):
        return i*j

    def dummy_s_function(qf_list):
        
        m = len(qf_list)
        for i in range(m):
            dummy_d_function(qf_list[i], qf_list[(i+1)%m])
        

    qf_list = [QuantumFloat(2) for j in range(3)]

    qf_list[-1][:] = 1

    qbool = QuantumBool()
    h(qbool)
    with control(qbool[0]):
        dummy_s_function(qf_list)
        print(qbool.qs)

    h(qbool)
    
    assert qbool.most_likely() == False


    #-------
    
    a = QuantumFloat(2, name = "a")
    b = QuantumFloat(2, name = "b")
    c = QuantumFloat(1, name = "c")
    d = QuantumFloat(2, name = "d")

    a[:] = 1
    b[:] = 1
    c[:] = 1

    def test_f(a):
        return a == 1

    from qrisp import ConditionEnvironment, ControlEnvironment
    
    with test_f(a):
        with b == c:
            x(d)
            print(d.qs)
    
    #Test if uncomputation worked properly
    assert len(a.qs.qv_list) == 4
    
    #Test if condition environments get invoked properly
    with b == 2:
        assert isinstance(b.qs.env_stack[-1], ConditionEnvironment)
        temp = (d == 1)
        # print(d.qs)
        with temp:
            assert isinstance(d.qs.env_stack[-1], ControlEnvironment)
            x(c)
            # print(c.qs)

    assert len(a.qs.qv_list) == 5
    
    #----------
    a = QuantumFloat(2, signed = True)
    b = QuantumFloat(2, signed = True)
    c = QuantumFloat(1, signed = True)
    d = QuantumFloat(2, signed = True)
    

    a[:] = 1
    b[:] = 1
    c[:] = 1

    def test_f(a):
        return a == 1

    with test_f(a):
        with b == c:
            x(d)
            print(d.qs)

    assert len(a.qs.qv_list) == 4
    


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


from qrisp import h, QuantumFloat, multi_measurement
import numpy as np

def test_comparison():
    
    def test_comparison_helper(qf_0, qf_1, comp):
        
        qf_0 = qf_0.duplicate()
        h(qf_0)
        
        if isinstance(qf_1, QuantumFloat):
            qf_1 = qf_1.duplicate()
            h(qf_1)
        
        if comp == "eq":
            qf_res = qf_0 == qf_1
        elif comp == "neq":
            qf_res = qf_0 != qf_1
        elif comp == "leq":
            qf_res = qf_0 <= qf_1
        elif comp == "geq":
            qf_res = qf_0 >= qf_1
        elif comp == "lt":
            qf_res = qf_0 < qf_1
        elif comp == "gt":
            qf_res = qf_0 > qf_1
        
        if isinstance(qf_1, QuantumFloat):
            mes_res = multi_measurement([qf_0, qf_1, qf_res])
            for (a, b, c) in mes_res.keys():
                print(a,b,c)
                if comp == "eq":
                    assert (a == b) == c
                elif comp == "neq":
                    assert (a != b) == c
                elif comp == "leq":
                    assert (a <= b) == c
                elif comp == "geq":
                    assert (a >= b) == c
                elif comp == "lt":
                    assert (a < b) == c
                elif comp == "gt":
                    assert (a > b) == c    
                    
            #Test correct phase behavior7
            statevector = qf_res.qs.statevector("array")
            angles = (np.angle(statevector[np.abs(statevector) > 1/2**((qf_0.size + qf_1.size)/2)]))
            assert np.sum(np.abs(angles)) < 0.1
        else:
            
            mes_res = multi_measurement([qf_0, qf_res])
            for (a, c) in mes_res.keys():
                b = qf_1
                print(a,b,c)
                if comp == "eq":
                    assert (a == b) == c
                elif comp == "neq":
                    assert (a != b) == c
                elif comp == "leq":
                    assert (a <= b) == c
                elif comp == "geq":
                    assert (a >= b) == c
                elif comp == "lt":
                    assert (a < b) == c
                elif comp == "gt":
                    assert (a > b) == c    
        
    a = QuantumFloat(5, 2, signed = False)
    b = QuantumFloat(5, -1, signed = True)
    #Test all the operators
    for comp in ["eq", "neq", "lt", "gt", "leq", "geq"]:
        test_comparison_helper(a, b, comp)
    
    #Test specific constellations of QuantumFloat parameters    
    a = QuantumFloat(3, -1, signed = True)
    b = QuantumFloat(5, 1, signed = True)
    test_comparison_helper(a, b, "lt")
    
    a = QuantumFloat(5, 2, signed = False)
    b = QuantumFloat(5, -1, signed = True)
    test_comparison_helper(a, b, "lt")
    
    a = QuantumFloat(4, -4, signed = False)
    b = QuantumFloat(4, 0, signed = True)
    test_comparison_helper(a, b, "lt")
    
    a = QuantumFloat(4, 1, signed = True)
    b = QuantumFloat(3, -1, signed = False)
    test_comparison_helper(a, b, "lt")
    
    #Test semi-classical comparisons
    a = QuantumFloat(4, signed = False)
    b = 4
    test_comparison_helper(a, b, "lt")
    test_comparison_helper(a, b, "gt")
    
    a = QuantumFloat(4, signed = True)
    b = 4
    test_comparison_helper(a, b, "lt")
    test_comparison_helper(a, b, "gt")
    
    a = QuantumFloat(4, -2, signed = False)
    b = 4
    test_comparison_helper(a, b, "lt")
    test_comparison_helper(a, b, "gt")
    
    a = QuantumFloat(8, 3, signed = False)
    b = 16
    test_comparison_helper(a, b, "lt")
    test_comparison_helper(a, b, "gt")



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

#
# Autogenerated by Thrift Compiler (0.15.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys
from .ttypes import *



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# -*- coding: utf-8 -*-


import numpy as np
from qrisp.circuit.operation import PTControlledOperation, ControlledOperation

#This function dissolves any Operation objects that have a definition circuit such that the result only consists of elementary gates
def transpile(qc, transpilation_level = np.inf, transpile_predicate = None, **kwargs):
    
    from qrisp.circuit import Operation, QuantumCircuit, Qubit, Clbit
    QuantumCircuit.fast_append = True
    from qrisp.circuit import Qubit, Clbit
    transpiled_qc = QuantumCircuit()
    
    # [transpiled_qc.add_qubit(Qubit(qb.identifier)) for qb in qc.qubits]
    # [transpiled_qc.add_clbit(Clbit(cb.identifier)) for cb in qc.clbits]
    
    for qb in qc.qubits:
        if isinstance(qb, Qubit):
            transpiled_qc.add_qubit(qb)
        else:
            transpiled_qc.add_qubit(Qubit(qb.identifier))
            
    for cb in qc.clbits:
        if isinstance(qb, Qubit):
            transpiled_qc.add_clbit(cb)
        else:
            transpiled_qc.add_clbit(Clbit(cb.identifier))
    
    for i in range(len(qc.data)):
        
        instruction = qc.data[i]
        
        if instruction.op.definition == None or transpilation_level <= 0:
            transpiled_qc.data.append(instruction)
        else:
            
            if transpile_predicate:
                if not transpile_predicate(instruction.op):
                    transpiled_qc.data.append(instruction)
                    continue
                
            definition_qc = transpile(instruction.op.definition, transpilation_level = transpilation_level -1, transpile_predicate = transpile_predicate).copy()
            
            
            translation_dic = {}
            
            translation_dic.update({definition_qc.qubits[j].identifier : instruction.qubits[j] for j in range(instruction.op.num_qubits)})
            translation_dic.update({definition_qc.clbits[j].identifier : instruction.clbits[j] for j in range(instruction.op.num_clbits)})
            
            extend(transpiled_qc, definition_qc, translation_dic)
    
    QuantumCircuit.fast_append = False
    
    
    if not kwargs or not hasattr(qc, "to_qiskit"):
        return transpiled_qc
    else:
        from qrisp import QuantumCircuit
        qiskit_qc = transpiled_qc.to_qiskit()
        
        from qiskit import transpile as qiskit_transpile
        
        transpiled_qiskit_qc = qiskit_transpile(qiskit_qc, **kwargs)
        
        qrisp_qc = QuantumCircuit.from_qiskit(transpiled_qiskit_qc)
        
        return qrisp_qc



    


def extend(qc_0, qc_1, translation_dic = "id"):
    
    if translation_dic == "id":
        
        translation_dic = {}
        for qb in qc_1.qubits:
            translation_dic[qb] = qb
            if not qb in qc_0.qubits:
                qc_0.add_qubit(qb)
            
        for cb in qc_1.clbits:
            translation_dic[qb] = qb
            if not cb in qc_0.clbits:
                qc_0.add_clbit(cb)
            

    #Copy in order to prevent modification
    translation_dic = dict(translation_dic)

    for key in list(translation_dic.keys()):
        
        if not isinstance(key, str):
            translation_dic[key.identifier] = translation_dic[key]


    for i in range(len(qc_1.data)):
        
        instruction_other = qc_1.data[i]
        qubits = []
        for qb in instruction_other.qubits:
            qubits.append(translation_dic[qb.identifier])
        
        clbits = []
        
        for cb in instruction_other.clbits:
            clbits.append(translation_dic[cb.identifier])
        
        instr_type = type(instruction_other)
        
        qc_0.data.append(instr_type(instruction_other.op, qubits, clbits))
        


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


import threading
from thrift.server import TServer

import logging
class StoppableThriftServer(TServer.TThreadedServer):
    
    def __init__(*args, **kwargs):
        args[0].is_running = False
        TServer.TThreadedServer.__init__(*args, **kwargs)
        
        
    def serve(self):
        
        self.serverTransport.listen()
        
        self.stop_flag = False
        thread_list = []
        while True:
            if self.stop_flag:
                for t in thread_list: t.join()
                break
            try:
                self.is_running = True
                client = self.serverTransport.accept()
                t = threading.Thread(target=self.handle, args=(client,))
                t.setDaemon(self.daemon)
                t.start()
                thread_list.append(t)
            except OSError:
                pass
            except KeyboardInterrupt:
                raise
            except Exception as x:
                logging.exception(x)
                
            

    def stop(self):
        self.stop_flag = True
        try:
            self.serverTransport.handle.shutdown(2)
        except OSError:
            pass
        self.serverTransport.handle.close()
        



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

import numpy as np
from qrisp.misc import int_as_array, int_encoder, array_as_int, gate_wrap
import sympy as sp

use_gray_code = False
try:
    from ffht import fht as fwht
    fwht(np.array([0,1,0,1]))
    #print("FALCONN fwht import successfull")
except:
    #print("FALCONN fwht import failed")
    
    def fwht(a):
        """In-place Fast Walsh–Hadamard Transform of array a."""
        h = 1
        while h < len(a):
            for i in range(0, len(a), h * 2):
                for j in range(i, i + h):
                    x = a[j]
                    y = a[j + h]
                    a[j] = x + y
                    a[j + h] = x - y
            h *= 2
    




def hamming_d(i,j, bit):
    return sum((int_as_array(i, bit) + int_as_array(j, bit))%2)

#Solution of the Hamming TSP
#Based on having two salesmen who always visit the place which is closest to them
#and which are reunified at the end of their travel
#The solution is now the path of the first one + the reversed path of the second
def hamming_tsp(location_list, bit_amount):
    if use_gray_code or len(location_list) == 2**(bit_amount-1) + 1:
        
        return gray_code(max(bit_amount-1,1))
    
    
    #print(location_list)
    if len(location_list) == 1:
        return list(location_list)
    
    if len(location_list) == 2:
        temp = list(location_list)
        temp.sort()
        #print([temp[1]] + hamming_movement(temp[1], temp[0], bit_amount))
        return [temp[1]]+ hamming_movement(temp[1], temp[0], bit_amount)
        
    
    #Copy list in order not to change the input
    location_list = list(location_list)
    
    #Create log-lists for the salesman pathes
    salesman_1 = [min(location_list)]
    location_list.pop(np.argmin(location_list))
    
    salesman_2 = [min(location_list)]
    location_list.pop(np.argmin(location_list))
    
    
    
    
    #Keep track of some stats
    empty_steps = 0
    
    while(len(location_list)):
        
        #Move first salesman
        
        
        #Find closest location
        best_distance = np.inf
        for i in range(len(location_list)):
            trial_distance = hamming_d(salesman_1[-1], location_list[i], bit_amount)#*len(location_list)
            #trial_distance += hamming_d(salesman_1[-1], salesman_2[-1], n)/len(location_list)
            
            if trial_distance < best_distance:
                best_distance = trial_distance
                best_candidate_index = i
        
        #Count steps which do not visit a relevant parity operator
        empty_steps += hamming_d(salesman_1[-1], location_list[best_candidate_index], bit_amount)-1
        
        #Remove target location from the location list
        target_location = location_list.pop(best_candidate_index)
        
        #Move salesman
        salesman_1 += hamming_movement(salesman_1[-1], target_location, bit_amount)
        
        #End procedure if the location list is empty now
        if len(location_list) == 0:
            break
        
        
        #Move second salesman (same steps as first salesman)
        
        best_distance = np.inf
        for i in range(len(location_list)):
            trial_distance = hamming_d(salesman_2[-1], location_list[i], bit_amount)#*len(location_list)
            #trial_distance += hamming_d(salesman_1[-1], location_list[i], n)#/len(location_list)
            
            if trial_distance < best_distance:
                best_distance = trial_distance
                best_candidate_index = i
                
        empty_steps += hamming_d(salesman_2[-1], location_list[best_candidate_index], bit_amount)-1
        target_location = location_list.pop(best_candidate_index)
        
        
        salesman_2 += hamming_movement(salesman_2[-1], target_location, bit_amount)
    
    
    #Unify salesmen
    salesman_1 += hamming_movement(salesman_1[-1], salesman_2[-1], bit_amount)
    
    salesman_2.pop(-1)
    
    #Construct solution (first salesman + second salesman reversed)
    solution = salesman_1 + salesman_2[::-1]
    
    #print("Reunion cost:")
    #print(hamming_d(salesman_1[-1], salesman_2[-1], n))
    #empty_steps += hamming_d(salesman_1[-1], salesman_2[-1], n) - 1
    #print("Percentage of empty steps")
    #print(empty_steps/parity_amount*100)
    #print(solution)
    
    
    if 2**max(bit_amount-1,1) < len(solution):
        return gray_code(max(bit_amount-1,1))
    
    
    return solution

def hamming_movement(init, target, bit_amount):
    salesman = [init]
    distance = hamming_d(init, target, bit_amount)
    while distance:
        distance_array = (int_as_array(salesman[-1], bit_amount) + int_as_array(target, bit_amount))%2
        movement_index = np.argmax(distance_array)
        movement_array = np.zeros(distance_array.shape)
        movement_array[movement_index] = 1
        
        salesman.append(array_as_int((int_as_array(salesman[-1],bit_amount) + movement_array)%2))
        distance = hamming_d(salesman[-1], target, bit_amount)
    #print((init, target))
    #print(salesman[1:])    
    return salesman[1:]



#Solves the hamming tsp problem ie. finds a path which visits each
#corner of the n-D unit cube exactly once and ends up in the origin again
def gray_code(n):
    #The key observation is that we can solve the problem for the n-D cube
    #if we solve it for the (n-1)-D cube such that the path starts at bin(0) and ends in bin(2**(n-2)).
    #We then concantenate this solution in the "lower half" of the n-D cube
    #with the reversed solution in the upper half applied
    
    #This concantenation will again give a solution which starts at 0 and ends in bin(2**((n+1)-2))
    #We thus have a recursion
    
    #Example for the 3D cube:
    #A possible solution for the 2D cube is [0,1,3,2]
    
    #The reversal is [2,3,1,0] which is [6,7,5,4] on the upper half
    #Concantenating gives [0,1,3,2,6,7,5,4] ie. a solution which starts in 0
    #and ends in 2**(3-1) = 4
    
    if n < 1:
        return []
    result_list = []
    
    
    if n == 1:
        return [1,0]
    
    temp_list = gray_code(n-1)
    
    result_list += [2**(n-1) + k for k in temp_list[::-1]]
    result_list += temp_list

    return result_list

#This function calculates the phases to put after each parity operator
#A given state will recieve then recieve the phase phi_i if it is true under
#the parity operator i. As each state is true on a unique combination of parity operators
#each state will recieve a unique phase. The total phase of any state is the sum of all the 
#parity operators it is True on. If we now want to get a certain constellation of total phases
#This gives us a system of linear equations, where the matrix is given by the parity matrix

#Target_phases is a list of phases , where the i-th entry describes the phase the i-state should pick up
#Note that the phases should be given as numbers between 0 and 2 where 2 describes a phase of 2pi
#This helps reducing the amount of float errors
def gen_phase_shifts(target_phases):
    
    
    try:
        temp = np.array(target_phases, dtype = np.longdouble)/(len(target_phases)/2)
    except TypeError:
        temp = [x/(len(target_phases)/2) for x in target_phases]
    
    #Transform
    fwht(temp)
    
    spectrum = temp
    
    return spectrum



#Solves the hamming tsp and turns it into a sequence of CNOT gates,
#all acting on the same qubit
#Returns a list of tuples indicating CNOT gates and the sequence of
#parity operators traversed
def single_qb_traversal(locations, bit_amount):
    
    
    #Solve Hamming TSP to get traversal sequence
    tr_seq = hamming_tsp(locations, bit_amount)
    
    #Make sure 0 appears last
    if tr_seq[-1] != 0:
        tr_seq = tr_seq + [0]
    
    if tr_seq[0] != 0:
        tr_seq = [0] + tr_seq
    
    #Turn into binaries
    tr_seq_bin = [int_as_array(x, bit_amount) for x in tr_seq]
    
    
    
    #Sequence of control bits.
    #The first control bit is the operation bit
    #This indicates later in the synthesis function, that
    #the unmodified parity operator has been loaded into the operation bit
    control_seq = [bit_amount-1]
    
    
    #Now successively calculate Hamming distance vector
    #and determine the non-zero coordinate
    temp = tr_seq[0]
    for i in range(1, len(tr_seq)):
        temp = (tr_seq_bin[i-1] + tr_seq_bin[i])%2
        temp = temp[::-1]
        
        control_seq.append(np.argmax(temp))
    
    return control_seq, tr_seq


#Similar to single_qb_traversal but this function also allows traversal
#of Parity operators in more than one qubit
#Returns a list of tuples indicating the applied CNOT gates
#and a list of integers indicating the sequence of traversed parity operators
def multi_qb_traversal(locations, bit_amount):
    
    #Copy locations list (in oder to prevent modification)
    
    locations = list(locations)
    
    #Sort list
    locations.sort()
    locations = locations[::-1]
    
    #Sort into a dictionary, where the keys are the operation qubits
    #and the values are list of parity operators that have to be traversed
    #on this qubit.
    single_qb_traversal_locations = {len(bin(x))-3: [] for x in locations}
    
    for x in locations:
        single_qb_traversal_locations[len(bin(x))-3].append(x)
    
    
    #Set up lists to be filled
    cnot_seq = []
    p_op_seq = []
    
    
    #Iterate through the qubits which are operated on
    for operation_bit in single_qb_traversal_locations.keys():
        
        #Convert the traversal locations into single qubit traversal routes
        #(for instance 4,5,7, is the traversal 0,1,3 on the qubit with significance 4)
        if operation_bit != 0:
            single_qb_locations = [x%2**(operation_bit) for x in single_qb_traversal_locations[operation_bit]]
        else:
            single_qb_locations = single_qb_traversal_locations[operation_bit]
        
        #Find single qubit traversal routes
        control_seq, p_op_temp = single_qb_traversal([0] + single_qb_locations, operation_bit+1)
        
        #Append to CNOT list
        cnot_seq += [(x,operation_bit) for x in control_seq]
        
        
        #Append to traversed parity operator list
        if operation_bit != 0:
            p_op_seq += [x + 2**(operation_bit) for x in p_op_temp]
        else:
            p_op_seq += p_op_temp
    
    return cnot_seq, p_op_seq


from qrisp.circuit.quantum_circuit import QuantumCircuit

compiled_gates = {}
compiled_pt_gates = {}

def gray_synth_gate(target_phases, phase_tolerant = False):
    
    target_phases_id = str(list(target_phases))
    
    if not phase_tolerant:
        if target_phases_id in compiled_gates:
            temp = compiled_gates[target_phases_id]
            return temp
    else:
        if target_phases_id in compiled_pt_gates:
            temp = compiled_pt_gates[target_phases_id]
            return temp

    bit_amount = int(np.log2(len(target_phases)))
    
    qc = QuantumCircuit(bit_amount)
    
    #Generate phase shifts
    phase_shifts = gen_phase_shifts(target_phases)
    
    #Set global phase to 0
    phase_shifts[0] = 0
    
    
    #Find out about non-zero parity operators that need to be traversed
    locations = []
    for i in range(len(phase_shifts)):
        try:
            if abs(phase_shifts[i]) < 1E-16:
                continue
        except TypeError:
            pass
        
        locations.append(i)
    
        
    #Generate CNOT list
    cnot_list, p_op_seq = multi_qb_traversal(locations, bit_amount)
    qb_list = qc.qubits
    
    #Apply CNOT list and add phases
    for i in range(len(cnot_list)):
            
        
        if (phase_tolerant == True and cnot_list[i][1] != bit_amount -1):
            #If we are in the phase tolerant mode we dont need to apply a phase here
            #because we are only interested in the phase DIFFERENCE of the 0 and 1 state of
            #the output qubit. As these phase shifts dont act on the output qubit they do not change
            #the phase difference and can therefore be ignored
            break
        
        #Set some aliases to keep the code readable
        operation_qubit = qb_list[cnot_list[i][1]]
        control_qubit = qb_list[cnot_list[i][0]]
        
        #Apply CNOT gate
        if cnot_list[i][0] != cnot_list[i][1]:
            control_qubit = qb_list[cnot_list[i][0]]
            qc.cx(control_qubit, operation_qubit)
        
        
        #Apply phase shifts
        if phase_shifts[p_op_seq[i]] == 0:
            continue
        qc.p(-phase_shifts[p_op_seq[i]], operation_qubit)
        phase_shifts[p_op_seq[i]] = 0
    
    
    from qrisp.core import reduce_depth
    
    if len(qc.qubits) < 8:
        try:
            qc = reduce_depth(qc)
        except TypeError:
            pass
    
    qc.qubits = qc.qubits[::-1]
    
    res = qc.to_gate("gray_phase_gate")
    res.target_phases = target_phases
    res.phase_tolerant = phase_tolerant
    
    if not phase_tolerant:
        compiled_gates[target_phases_id] = res
    else:
        compiled_pt_gates[target_phases_id] = res
    
    
    res.permeability = {i : True for i in range(len(qc.qubits))}
    res.is_qfree = True
    
    return res
    


#Function apply gray synthesis to quantum variable qv

#Target_phases is a list of phases , where the i-th entry describes the phase the i-state should pick up
#Note that the phases should be given as numbers between 0 and 2 where 2 describes a phase of 2pi
#This helps reducing the amount of float errors
@gate_wrap(permeability = "full", is_qfree = True)
def gray_phase_synth(qv, target_phases, phase_tolerant = False):
    qv.qs.append(gray_synth_gate(target_phases, phase_tolerant), qv.reg)
    

#Similar function as the above but takes a QuantumCircuit and a qubit list as arguments instead
def gray_phase_synth_qb_list(qc, qb_list, target_phases, phase_tolerant = False):
    qc.append(gray_synth_gate(target_phases, phase_tolerant), qb_list)

    

    
#Function to use gray synthesis for logic synthesis
#The general idea is to apply gray synthesis to the input variable and the output variables
#The output variable however is enclosed H gates which means that it will be |1> if we synthesized
#the |-> state and |0> if we synthesized the |+> state
#We choose the phases for the states which have a 0 in the output variable to be 0
#and the phases for the states which have a 1 in the output variable according to wether
#the truth table to be synthesized requires a 1 or a 0

#An additional perk of this function is, that it support phase tolerant synthesis
#This means that we are tolerant regarding the phase the out will have
#ie. if the truth table for a given state says 1 and we activate the phase_tolerant option
#we will get the state exp(i*phi)*|1> so a one but with a phase which might vary
#for different input states. This tolerance regarding the phase allows us to skip half
#of the synthesis procedure making it very cheap regarding the gate count

#input_var is the quantum variable which contains the input
#output_var is the quantum variable where the output should be synthesized
#Note that output_var doesn't need to be of size one - the next parameter
#qb_nr indicates on which qubit of output_var the synthesis should take place
#Finally tt is the truth table to be synthesized
@gate_wrap(is_qfree = True, permeability = [0])
def gray_logic_synth_single_qb(input_var, output_var, qb_nr, tt, phase_tolerant = False):
    if input_var.size != tt.bit_amount:
        raise Exception("Input variable does not contain enough qubits to encode truth table")
    
    if tt.shape[1] != 1:
        raise Exception("This function can only encode single column truth tables")
    
    
    #Generate target phases such that all the states with a 0 in the output qubit
    #get phase 0 and the states with a 1 in the input qubit get a phases of pi
    #if the truth table says 1 or a phase of 0 otherwise
    target_phases = np.array(tt.shape[0]*[0] + list(tt.n_rep[:,0]))*np.pi
    
    #Apply h gate to the output qubit in order to get logic states from phases
    from qrisp import h
    h(output_var[qb_nr])
    input_var.qs.append(gray_synth_gate(target_phases, phase_tolerant), (input_var.reg + [output_var[qb_nr]])[::-1])
    h(output_var[qb_nr])
    return
    
    


#This function uses it's single qubit version iteratively to synthesize
#truth tables with more than one column
@gate_wrap(is_qfree = True, permeability = [0])
def gray_logic_synth(input_var, output_var, tt, phase_tolerant = False, lin_solve = False):
    
    
    
    if len(input_var) != tt.bit_amount:
        raise Exception("Input variable does not contain enough qubits to encode truth table")
    
    if len(output_var) != tt.shape[1]:
        raise Exception("output variable does not contain enough qubits to encode truth table")
    
    input_var_dupl = input_var.duplicate()
    output_var_dupl = output_var.duplicate(qs = input_var_dupl.qs)
    
    #Iterate throgh all the truth table columns and synthesize each one separately
    
    residual_phases = np.zeros(tt.shape[0])
    
    for i in range(tt.shape[1]):
        gray_logic_synth_single_qb(input_var_dupl, output_var_dupl, i, tt.sub_table(i), phase_tolerant = True)

        temp = (tt.sub_table(i).n_rep - 1/2)*np.pi/2
        temp = temp.transpose()[0]
        
        residual_phases += temp
    
    from qrisp.core import reduce_depth
    
    qc = input_var_dupl.qs.transpile()
    
    if len(qc.qubits) < 10 and input_var.size != 1:
        qc = reduce_depth(qc)
    
    input_var_dupl.qs.data = []
    
    if not phase_tolerant:
        input_var_dupl.reg = input_var_dupl.reg[::-1]
        
        gray_phase_synth(input_var_dupl, residual_phases)
        
        input_var_dupl.reg = input_var_dupl.reg[::-1]
    
    qc.data.extend(input_var_dupl.qs.data)
    
    input_var.qs.append(qc.to_gate(), input_var.reg + output_var.reg)
    
#This function uses it's single qubit version iteratively to synthesize
#truth tables with more than one column
@gate_wrap
def gray_logic_synth_qb_list(input_qb_list, output_qb, qs, tt, phase_tolerant = False, inverse = False, gate_name = None):
    
    from qrisp import QuantumSession, QuantumVariable
    qs_temp = QuantumSession()
    
    input_var = QuantumVariable(len(input_qb_list), qs_temp)
    output_var = QuantumVariable(1, qs_temp)
    
    gray_logic_synth(input_var, output_var, tt, phase_tolerant)
    
    
    temp_gate = qs_temp.to_gate()
    
    if gate_name != None:
        temp_gate.name = gate_name
        
    if inverse:
        qs.append(temp_gate.inverse(), input_qb_list + [output_qb])
    else:
        qs.append(temp_gate, input_qb_list + [output_qb])
        

       


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 06.05.2022
# import pytest

from qrisp import QuantumString

def test_hello_world():
    q_str = QuantumString(size = len("hello world"))
    q_str.encode("hello world")

    assert list(q_str.get_measurement().keys())[0] == "hello world"

